/*****************************************************************************/
/*
                                 Error.c

Error reporting functions.

All errors are reported using HTML formatting.  The assumptions being: 

  o  That most errors will occur before a document begins being transfered
     (e.g. file not found), and so its content type has not been set, and 
     the error can sent with a "text/html" type.

  o  HTML documents are probably more extensive than plain-text documents.
     Hence if an error occurs during transfer (e.g. record too big for user's
     buffer) it can be reported within the document with correct formatting 
     (document context notwithstanding).

  o  If an error occurs while transfering a plain-text document the HTML
     message is still readable, albeit a little cluttered.  Additional,
     obvious text is included to try and highlight the message.

  o  In non-"text/..." transfers the message text is not included anyway.

After a request thread is created and an error is generated the storage
'rqptr->ErrorReportPtr' becomes non-NULL.  This mechanism can be used 
to detect whether an error has occured.  Most common errors create an error 
message in heap memory and point at this using 'rqptr->ErrorReportPtr'. 
This message can be output at some time convenient to the task underway.  
After an error message is sent to the client the 'rqptr->ErrorReportPtr' 
is returned to a NULL value to indicate no outstanding error (but usually the 
thread is disposed of after an error occurs).  Notification of other, more 
critical errors are sent directly to the client at the time they occur (for 
instance, failed heap memory allocation).

If ErrorGeneral() is called without first setting rqptr->ResponseStatusCode
it will default to 500, or "internal server error".  Therefore string
overflows and the like can be called without setting this code.  Query string
errors should set to 403 or 404 indication request cannot be processed.

If ErrorVmsStatus() is called without first setting rqptr->ResponseStatusCode
the function will attempt to determine the most appropriate HTTP status code
from the VMS status value.  It will then default to 500, or "internal server
error", indicating some unsual (e.g. non file system) condition.

The server configuration can supply additional information included whenever
an error report is generated.  This includes a URL for contacting the system
administrator (note: URL, not text, etc.) when an implementation,
configuration, process quota, etc., error is reported.   It also can provide
text to be included against any other error report.  This text can contain
HTML markup, carriage control, etc., and is designed for a brief message
containing a link to an explanation page.

Error reporting can be performed by local (server-internal) redirection.  This
provides the opportunity for greater local-site control over content and
format.  When an error is redirected essential error information is encoded
into a query-string formatted (URL-encoded) string.  The redirection merely
creates a new request with the redirection document (SSI or CGI script) as the
path and this information as the query string.  The reporting document
environment can then use this information via the following CGI variables:

FORM_ERROR_ABOUT       item message is about (if applicable)
FORM_ERROR_ABOUT2      additional information (if applicable)
FORM_ERROR_BASIC       basic authentication challenge
FORM_ERROR_DIGEST      digest authentication challenge
FORM_ERROR_LINE        name of source code line (e.g. "1732")
FORM_ERROR_MODULE      name of source code module (e.g. "REQUEST")
FORM_ERROR_REALM       authentication realm
FORM_ERROR_STATUS      HTTP status code (e.g. "404")
FORM_ERROR_TEXT        server-generated error message
FORM_ERROR_TEXT2       server suggestion about what to do about it :^)
FORM_ERROR_VMS         VMS status value (if applicable)


VERSION HISTORY
---------------
02-JAN-99  MGD  proxy service
18-OCT-98  MGD  error report redirection
30-AUG-97  MGD  401 without a realm is now converted into a 403
09-AUG-97  MGD  message database, with considerable rewrite!
01-FEB-97  MGD  HTTPd version 4
06-JUN-96  MGD  added authentication failure
01-DEC-95  MGD  HTTPd version 3
25-MAR-95  MGD  minor functional and cosmetic changes
20-DEC-94  MGD  multi-threaded daemon
*/
/*****************************************************************************/

/* standard C header files */
#include <ctype.h>
#include <stdio.h>
#include <string.h>

/* VMS related header files */
#include <descrip.h>
#include <iodef.h>
#include <ssdef.h>
#include <stsdef.h>

/* application related header files */
#include "wasd.h"

#define WASD_MODULE "ERROR"

/******************/
/* global storage */
/******************/

$DESCRIPTOR (ErrorNumberFaoDsc, "!UL\0");
$DESCRIPTOR (ErrorVmsStatusFaoDsc, "%X!8XL\0");
char  ErrorScratchString [16];
$DESCRIPTOR (ErrorScratchStringDsc, ErrorScratchString);

char  ErrorSanityCheck [] = "sanity check failure!",
      ErrorServerFao [] = "significant HTTPd sys$fao() failed.",
      ErrorFormAbout [] = "&error_about=",
      ErrorFormAbout2 [] = "&error_about2=",
      ErrorFormBasic [] = "&error_basic=",
      ErrorFormDigest [] = "&error_digest=",
      ErrorFormHttpd [] = "?httpd=error",
      ErrorFormLine [] = "&error_line=",
      ErrorFormModule [] = "&error_module=",
      ErrorFormProxy [] = "&error_proxy=",
      ErrorFormRealm [] = "&error_realm=",
      ErrorFormStatus [] = "&error_status=",
      ErrorFormText [] = "&error_text=",
      ErrorFormText2 [] = "&error_text2=",
      ErrorFormVms [] = "&error_vms=";

char  *ErrorRecommendNotifyPtr;

/********************/
/* external storage */
/********************/

#ifdef DBUG
extern boolean Debug;
#else
#define Debug 0 
#endif

extern boolean  LoggingEnabled;
extern int  AdminWatchEnabled;
extern char  HttpProtocol[];
extern char  SoftwareID[];
extern char  Utility[];
extern struct AccountingStruct  Accounting;
extern struct ConfigStruct  Config;
extern struct MsgStruct  Msgs;

/*****************************************************************************/
/*
Exit the application after output of the module and line location reporting 
the error, a brief explanation, and then exiting generating a DCL-reported 
message.
*/

ErrorExitVmsStatus
(
int VmsStatus,
char *Explanation,
char *SourceModuleName,
int SourceLineNumber
)
{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ErrorExitVmsStatus()\n");

   fprintf (stdout,
"%%%s-E-SOFTWAREID, %s\n\
-%s-E-WHERE, module: %s line: %d\n\
-%s-E-WHAT, %s\n",
            Utility, SoftwareID,
            Utility, SourceModuleName, SourceLineNumber,
            Utility, Explanation);

   if (!VmsStatus) VmsStatus = STS$K_ERROR | STS$M_INHIB_MSG;
   exit (VmsStatus);
}

/****************************************************************************/
/*
It is a fatal error to call this function without an error message.  If a
response header and output has already been generated insert a plain-text and
HTML obvious divider into the output (hopefully it's text ... it doesn't really
matter anyway, it's an error condition after all :^) and then send the actual
error message.  If no response has yet been made then either redirect error
processing appropriately (using the query-string format error information
generated by ErrorVmsStatus() and ErrorGeneral()), or generate an response
header then send the message.
*/

ErrorSendToClient (struct RequestStruct *rqptr)

{
   /* contains HTML, also enough to make it obvious in plain text */
   static $DESCRIPTOR (DividerFaoDsc,
"<FONT COLOR=\"#ff0000\"><B><HR>\n\
\n\
<!!-- -- --   !AZ   -- -- -->\n\
\n");

   static $DESCRIPTOR (BufferDsc, "");
   static $DESCRIPTOR (ReportPathFaoDsc, "");

   register char  *cptr, *sptr;
   register unsigned long  *vecptr;

   int  status,
        Count;
   unsigned short  Length;
   unsigned long  FaoVector [32];
   char  Buffer [1024];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ErrorSendToClient()\n");

   if (rqptr->ErrorReportPtr == NULL)
      ErrorExitVmsStatus (0, ErrorSanityCheck, FI_LI);

   /* shouldn't happen ... but */
   if (!rqptr->ResponseStatusCode) rqptr->ResponseStatusCode = 500;

   if (rqptr->ResponseHeaderPtr != NULL &&
       rqptr->OutputBufferPtr != NULL)
   {
      /*******************/
      /* provide divider */
      /*******************/

      BufferDsc.dsc$a_pointer = Buffer;
      BufferDsc.dsc$w_length = sizeof(Buffer)-1;

      status = sys$fao (&DividerFaoDsc, &Length, &BufferDsc,
                        MsgFor(rqptr,MSG_STATUS_ERROR));
      if (VMSnok (status) || status == SS$_BUFFEROVF)
         ErrorExitVmsStatus (status, ErrorServerFao, FI_LI);
      Buffer[Length] = '\0';

      NetWrite (rqptr, &ErrorMessageDividerAst, Buffer, Length);

      return;
   }

   if (rqptr->ErrorReportByRedirect)
   {
      /******************************/
      /* redirect to error reporter */
      /******************************/

      /* log the original request */
      if (LoggingEnabled) Logging (rqptr, LOGGING_ENTRY);

      BufferDsc.dsc$a_pointer = Buffer;
      BufferDsc.dsc$w_length = sizeof(Buffer)-1;

      ReportPathFaoDsc.dsc$a_pointer = Config.ErrorReportPath;
      ReportPathFaoDsc.dsc$w_length = Config.ErrorReportPathLength;

      status = sys$fao (&ReportPathFaoDsc, &Length, &BufferDsc,
                        rqptr->ResponseStatusCode);
      if (VMSnok (status) || status == SS$_BUFFEROVF)
         ErrorExitVmsStatus (status, ErrorServerFao, FI_LI);
      Buffer[Length] = '\0';

      Count = Length;
      Count += rqptr->ErrorReportLength;
      Count++;

      rqptr->LocationPtr = sptr = VmGetHeap (rqptr, Count);

      for (cptr = Buffer; *cptr; *sptr++ = *cptr++);
      for (cptr = rqptr->ErrorReportPtr; *cptr; *sptr++ = *cptr++);
      *sptr = '\0';

      /* indicate the error message has been sent */
      rqptr->ErrorReportPtr = NULL;
      rqptr->ErrorReportLength = 0;

      /* indicate error redirection in this thread-permanent storage */
      rqptr->ErrorReportStatusCode = rqptr->ResponseStatusCode;
      rqptr->ErrorReportAuthRealmDescriptPtr = rqptr->AuthRealmDescriptionPtr;

      /* errors are always sent as the request is finalized, continue that */
      RequestEnd (rqptr);

      return;
   }

   /***********************/
   /* provide HTTP header */
   /***********************/

   HttpHeader (rqptr, rqptr->ResponseStatusCode, "text/html", -1, NULL, NULL);

   /* errors are always sent as the request is finalized, continue that */
   NetWrite (rqptr, &RequestEnd,
             rqptr->ErrorReportPtr, rqptr->ErrorReportLength);

   /* indicate the error message has been sent */
   rqptr->ErrorReportPtr = NULL;
   rqptr->ErrorReportLength = 0;
}

/*****************************************************************************/
/*
Divider delivered, now output the error report.
*/

ErrorMessageDividerAst (struct RequestStruct *rqptr)

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ErrorMessageDividerAst()\n");

   /* errors are always sent as the request is finalized, continue that */
   NetWrite (rqptr, &RequestEnd,
             rqptr->ErrorReportPtr, rqptr->ErrorReportLength);

   /* indicate the error message has been sent */
   rqptr->ErrorReportPtr = NULL;
   rqptr->ErrorReportLength = 0;
}

/*****************************************************************************/
/*
Generate an error message about a VMS status problem for subsequent reporting
to the client.  Generally most errors are reported as documents in their own
right, making the full HTML document desirable.  If reported part-way through
some action the redundant tags should be ignored anyway.  If error redirection
applies to the request then generate the error information in a form suitable
for provision as a request query-string.
*/

ErrorVmsStatus
(
struct RequestStruct *rqptr,
int StatusValue,
char *SourceModuleName,
int SourceLineNumber
)
{
   static $DESCRIPTOR (ErrorFaoDsc,
"<HTML>\n\
<HEAD>\n\
!AZ\
<TITLE>!AZ !UL</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1>!AZ!!</H1>\n\
<P>!AZ\n\
<P>!AZ ... <TT>!AZ</TT>\n\
<!!-- sts: %X!XL \"!AZ\" -->\n\
!AZ!AZ\
!AZ\
</BODY>\n\
</HTML>\n");

   register char  *cptr, *sptr, *zptr;
   register unsigned long  *vecptr;

   int  status;
   unsigned short  Length;
   unsigned long  FaoVector [32];
   char  *ContentTypePtr,
         *ExplanationPtr,
         *RecommendLfPtr,
         *RecommendPtr;
   char  Message [256],
         Buffer [4096];
   $DESCRIPTOR (MessageDsc, Message);
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "ErrorVmsStatus() %%X%08.08X %s %d\n",
               StatusValue, SourceModuleName, SourceLineNumber);
                                 
   /* don't overwrite any existing error message */
   if (rqptr->ErrorReportPtr != NULL) return;

   /* if this request is processing an error report redirect */
   if (rqptr->ErrorReportByRedirect &&
       rqptr->ErrorReportStatusCode)
   {
      /* problem processing the error report, no infinite loops here please! */
      rqptr->ErrorReportByRedirect = false;
      rqptr->ErrorReportStatusCode = 0;
      rqptr->ErrorReportAuthRealmDescriptPtr = NULL;
   }

   if (rqptr->ResponseHeaderPtr != NULL &&
       rqptr->OutputBufferPtr != NULL)
   {
      /* some output has already occured, internal error reporting */
      rqptr->ErrorReportByRedirect = false;
   }

   if (rqptr->ResponseStatusCode == 401) HttpHeaderChallenge (rqptr);

   if (rqptr->ContentType.ContentTypePtr == NULL)
      ContentTypePtr = "";
   else
      ContentTypePtr = rqptr->ContentType.ContentTypePtr;

   if (StatusValue == RMS$_FNF)
   {
      if (!rqptr->ResponseStatusCode) rqptr->ResponseStatusCode = 404;
      if (ConfigSameContentType (ContentTypePtr, "text/", 5))
         ExplanationPtr = MsgFor(rqptr,MSG_STATUS_DOC_NOT_FOUND);
      else
         ExplanationPtr = MsgFor(rqptr,MSG_STATUS_FILE_NOT_FOUND);
   }
   else
   if (StatusValue == RMS$_PRV)
   {
      if (!rqptr->ResponseStatusCode) rqptr->ResponseStatusCode = 403;
      if (ConfigSameContentType (ContentTypePtr, "text/", 5))
         ExplanationPtr = MsgFor(rqptr,MSG_STATUS_DOC_PROTECTION);
      else
         ExplanationPtr = MsgFor(rqptr,MSG_STATUS_FILE_PROTECTION);
   }
   else
   if (VMSok (status = sys$getmsg (StatusValue, &Length, &MessageDsc, 1, 0)))
   {
      if (!rqptr->ResponseStatusCode)
      {
         if (StatusValue == RMS$_DNF ||
             StatusValue == RMS$_SYN ||
             StatusValue == RMS$_DEV ||
             StatusValue == RMS$_FLK ||
             StatusValue == RMS$_FNM ||
             StatusValue == RMS$_TYP ||
             StatusValue == RMS$_DIR ||
             StatusValue == SS$_NOSUCHFILE ||
             StatusValue == SS$_DIRNOTEMPTY ||
             StatusValue == SS$_EXDISKQUOTA)
            rqptr->ResponseStatusCode = 404;
         else
         if (StatusValue == SS$_NOPRIV)
            rqptr->ResponseStatusCode = 403;
         else
            rqptr->ResponseStatusCode = 500;
      }
      Message[Length] = '\0';
      cptr = sptr = ExplanationPtr = Message;
      *cptr = toupper(*cptr);

      /* don't allow access violation to be confused with forbidden access! */
      if (StatusValue == SS$_ACCVIO)
      {
         while (*cptr && *cptr != ',') cptr++;
         if (*cptr) strcpy (cptr, " (segmentation fault)");
         cptr = sptr;
      }

      /* improve the look by removing any embedded sys$fao() formatting */
      while (*cptr)
      {
         if (*cptr == '!')
         {
            cptr++;
            *sptr++ = '?';
            /* step over any field width digits */
            while (isdigit(*cptr)) cptr++;
            /* usually two formatting characters */
            if (isalpha(*cptr)) cptr++;
            if (isalpha(*cptr)) cptr++;
         }
         else
            *sptr++ = *cptr++;
      }
      *sptr = '\0';
   }
   else
      strcpy (Message, "*ERROR* sys$getmsg()");

   if (!rqptr->ResponseStatusCode) rqptr->ResponseStatusCode = 500;

   if (rqptr->ErrorTextPtr == NULL)
   {
      if (rqptr->PathInfoPtr == NULL)
         rqptr->ErrorTextPtr = MsgFor(rqptr,MSG_STATUS_NO_INFORMATION);
      else
         rqptr->ErrorTextPtr = rqptr->PathInfoPtr;
   }

   if (rqptr->ErrorOtherTextPtr == NULL || !Config.IncludeCommentedInfo)
      rqptr->ErrorOtherTextPtr = MsgFor(rqptr,MSG_STATUS_NO_INFORMATION);

   RecommendPtr = NULL;
   if (!Config.ErrorRecommend || rqptr->ResponseStatusCode == 401)
      RecommendPtr = "";
   else
   if (StatusValue == RMS$_FNF ||
       StatusValue == RMS$_DNF ||
       StatusValue == SS$_NOSUCHFILE)
      RecommendPtr = MsgFor(rqptr,MSG_STATUS_ADVISE_NOSUCHFILE);
   else
   if (StatusValue == RMS$_PRV)
      RecommendPtr = MsgFor(rqptr,MSG_STATUS_ADVISE_PRV);
   else
   if (StatusValue == SS$_NOPRIV)
      RecommendPtr = MsgFor(rqptr,MSG_STATUS_ADVISE_NOPRIV);
   else
   if (StatusValue == RMS$_SYN ||
       StatusValue == RMS$_DEV ||
       StatusValue == RMS$_DIR ||
       StatusValue == RMS$_FNM ||
       StatusValue == RMS$_TYP)
      RecommendPtr = MsgFor(rqptr,MSG_STATUS_ADVISE_SYNTAX);
   else
   if (StatusValue == RMS$_FLK)
      RecommendPtr = MsgFor(rqptr,MSG_STATUS_ADVISE_FLK);
   else
   if (StatusValue == SS$_DIRNOTEMPTY ||
       StatusValue == SS$_EXDISKQUOTA)
      RecommendPtr = MsgFor(rqptr,MSG_STATUS_ADVISE_CORRECT);

   if (RecommendPtr == NULL)
      RecommendPtr = RecommendLfPtr = "";
   else
      RecommendLfPtr = "\n";

   if (rqptr->AdminWatchItem &&
       AdminWatchEnabled && (AdminWatchEnabled & ADMIN_WATCH_ERROR))
      AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_ERROR,
                      "!AZ:!UL !3ZL %X!8XL \"!AZ\" !AZ !AZ",
                      SourceModuleName, SourceLineNumber,
                      rqptr->ResponseStatusCode, StatusValue,
                      ExplanationPtr, rqptr->ErrorTextPtr,
                      rqptr->ErrorOtherTextPtr);

   if (rqptr->ErrorReportByRedirect)
   {
      /******************************************/
      /* redirect report, generate query string */
      /******************************************/

      zptr = (sptr = Buffer) + sizeof(Buffer);

      for (cptr = ErrorFormHttpd; *cptr && sptr < zptr; *sptr++ = *cptr++);

      sys$fao (&ErrorNumberFaoDsc, 0, &ErrorScratchStringDsc,
               rqptr->ResponseStatusCode);
      for (cptr = ErrorFormStatus; *cptr && sptr < zptr; *sptr++ = *cptr++);
      for (cptr = ErrorScratchString; *cptr && sptr < zptr; *sptr++ = *cptr++);

      sys$fao (&ErrorVmsStatusFaoDsc, 0, &ErrorScratchStringDsc,
               StatusValue);
      for (cptr = ErrorFormVms; *cptr && sptr < zptr; *sptr++ = *cptr++);
      sptr += UrlEncodeString (ErrorScratchString, sptr, zptr-sptr);

      for (cptr = ErrorFormAbout; *cptr && sptr < zptr; *sptr++ = *cptr++);
      sptr += UrlEncodeString (rqptr->ErrorTextPtr, sptr, zptr-sptr);

      for (cptr = ErrorFormAbout2; *cptr && sptr < zptr; *sptr++ = *cptr++);
      sptr += UrlEncodeString (rqptr->ErrorOtherTextPtr, sptr, zptr-sptr);

      for (cptr = ErrorFormText; *cptr && sptr < zptr; *sptr++ = *cptr++);
      sptr += UrlEncodeString (ExplanationPtr, sptr, zptr-sptr);

      for (cptr = ErrorFormText2; *cptr && sptr < zptr; *sptr++ = *cptr++);
      sptr += UrlEncodeString (RecommendPtr, sptr, zptr-sptr);

      for (cptr = ErrorFormModule; *cptr && sptr < zptr; *sptr++ = *cptr++);
      for (cptr = SourceModuleName; *cptr && sptr < zptr; *sptr++ = *cptr++);

      sys$fao (&ErrorNumberFaoDsc, 0, &ErrorScratchStringDsc,
               SourceLineNumber);
      for (cptr = ErrorFormLine; *cptr && sptr < zptr; *sptr++ = *cptr++);
      for (cptr = ErrorScratchString; *cptr && sptr < zptr; *sptr++ = *cptr++);

      for (cptr = ErrorFormProxy; *cptr && sptr < zptr; *sptr++ = *cptr++);
      if (rqptr->ProxyTaskPtr != NULL)
         for (cptr = rqptr->ServicePtr->ServerHostPort;
              *cptr && sptr < zptr;
              *sptr++ = *cptr++);

      if (rqptr->ResponseStatusCode == 401)
      {
         for (cptr = ErrorFormRealm; *cptr && sptr < zptr; *sptr++ = *cptr++);
         sptr += UrlEncodeString (rqptr->AuthRealmPtr, sptr, zptr-sptr);

         if (rqptr->AuthDigestChallengePtr != NULL)
            if (rqptr->AuthDigestChallengePtr[0])
            {
               for (cptr = ErrorFormDigest;
                    *cptr && sptr < zptr;
                    *sptr++ = *cptr++);
               cptr = rqptr->AuthDigestChallengePtr;
               sptr += UrlEncodeString (cptr, sptr, zptr-sptr);
            }

         if (rqptr->AuthBasicChallengePtr != NULL)
            if (rqptr->AuthBasicChallengePtr[0])
            {
               for (cptr = ErrorFormBasic;
                    *cptr && sptr < zptr;
                    *sptr++ = *cptr++);
               cptr = rqptr->AuthBasicChallengePtr;
               sptr += UrlEncodeString (cptr, sptr, zptr-sptr);
            }
      }

      *sptr++ = '\0';
      rqptr->ErrorReportLength = Length = sptr - Buffer;
      rqptr->ErrorReportPtr = VmGetHeap (rqptr, Length);
      memcpy (rqptr->ErrorReportPtr, Buffer, Length);

      return;
   }

   /**********************************/
   /* generate internal error report */
   /**********************************/

   vecptr = FaoVector;
   *vecptr++ = ErrorSourceInfo (SourceModuleName, SourceLineNumber);
   *vecptr++ = cptr = MsgFor(rqptr,MSG_STATUS_ERROR);
   *vecptr++ = rqptr->ResponseStatusCode;
   *vecptr++ = cptr;
   if (rqptr->ProxyTaskPtr != NULL)
      *vecptr++ = MsgFor(rqptr,MSG_STATUS_REPORTED_BY_PROXY);
   else
      *vecptr++ = MsgFor(rqptr,MSG_STATUS_REPORTED_BY_SERVER);
   *vecptr++ = ExplanationPtr;
   *vecptr++ = rqptr->ErrorTextPtr;
   *vecptr++ = StatusValue;
   *vecptr++ = rqptr->ErrorOtherTextPtr;
   *vecptr++ = RecommendPtr;
   *vecptr++ = RecommendLfPtr;
   *vecptr++ = MsgFor(rqptr,MSG_STATUS_INFO);

   status = sys$faol (&ErrorFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
      ErrorExitVmsStatus (status, ErrorServerFao, FI_LI);

   Buffer[rqptr->ErrorReportLength = Length] = '\0';
   rqptr->ErrorReportPtr = VmGetHeap (rqptr, Length+1);
   memcpy (rqptr->ErrorReportPtr, Buffer, Length);
   rqptr->ErrorReportPtr[rqptr->ErrorReportLength = Length] = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", rqptr->ErrorReportPtr);
}

/*****************************************************************************/
/*
Generate an error message about a general (non-VMS status) problem for
subsequent reporting to the client.  If error redirection applies to the
request then generate the error information in a form suitable for provision as
a request query-string.
*/
 
ErrorGeneral
(
struct RequestStruct *rqptr,
char *ExplanationPtr,
char *SourceModuleName,
int SourceLineNumber
)
{
   static $DESCRIPTOR (ErrorFaoDsc,
"<HTML>\n\
<HEAD>\n\
!AZ\
<TITLE>!AZ !UL</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1>!AZ!!</H1>\n\
<P>!AZ\n\
<P>!AZ\n\
!AZ\
</BODY>\n\
</HTML>\n");

   register char  *cptr, *sptr, *zptr;
   register unsigned long  *vecptr;

   int  status;
   unsigned short  Length;
   unsigned long  FaoVector [32];
   char  Buffer [4096];
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ErrorGeneral()\n");

   /* don't overwrite any existing error message */
   if (rqptr->ErrorReportPtr != NULL) return;

   /* if this request is processing an error report redirect */
   if (rqptr->ErrorReportByRedirect &&
       rqptr->ErrorReportStatusCode)
   {
      /* problem processing the error report, no infinite loops here please! */
      rqptr->ErrorReportByRedirect = false;
      rqptr->ErrorReportStatusCode = 0;
      rqptr->ErrorReportAuthRealmDescriptPtr = NULL;
   }

   if (rqptr->ResponseHeaderPtr != NULL &&
       rqptr->OutputBufferPtr != NULL)
   {
      /* some output has already occured, internal error reporting */
      rqptr->ErrorReportByRedirect = false;
   }

   if (rqptr->ResponseStatusCode == 401) HttpHeaderChallenge (rqptr);

   if (!rqptr->ResponseStatusCode) rqptr->ResponseStatusCode = 500;

   if (rqptr->AdminWatchItem &&
       AdminWatchEnabled && (AdminWatchEnabled & ADMIN_WATCH_ERROR))
      AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_ERROR,
                      "!AZ:!UL !3ZL \"!AZ\"",
                      SourceModuleName, SourceLineNumber,
                      rqptr->ResponseStatusCode, ExplanationPtr);

   if (rqptr->ErrorReportByRedirect)
   {
      /******************************************/
      /* redirect report, generate query string */
      /******************************************/

      zptr = (sptr = Buffer) + sizeof(Buffer);

      for (cptr = ErrorFormHttpd; *cptr && sptr < zptr; *sptr++ = *cptr++);

      sys$fao (&ErrorNumberFaoDsc, 0, &ErrorScratchStringDsc,
               rqptr->ResponseStatusCode);
      for (cptr = ErrorFormStatus; *cptr && sptr < zptr; *sptr++ = *cptr++);
      for (cptr = ErrorScratchString; *cptr && sptr < zptr; *sptr++ = *cptr++);

      for (cptr = ErrorFormText; *cptr && sptr < zptr; *sptr++ = *cptr++);
      sptr += UrlEncodeString (ExplanationPtr, sptr, zptr-sptr);

      for (cptr = ErrorFormModule; *cptr && sptr < zptr; *sptr++ = *cptr++);
      for (cptr = SourceModuleName; *cptr && sptr < zptr; *sptr++ = *cptr++);

      sys$fao (&ErrorNumberFaoDsc, 0, &ErrorScratchStringDsc,
               SourceLineNumber);
      for (cptr = ErrorFormLine; *cptr && sptr < zptr; *sptr++ = *cptr++);
      for (cptr = ErrorScratchString; *cptr && sptr < zptr; *sptr++ = *cptr++);

      for (cptr = ErrorFormProxy; *cptr && sptr < zptr; *sptr++ = *cptr++);
      if (rqptr->ProxyTaskPtr != NULL)
         for (cptr = rqptr->ServicePtr->ServerHostPort;
              *cptr && sptr < zptr;
              *sptr++ = *cptr++);

      if (rqptr->ResponseStatusCode == 401)
      {
         for (cptr = ErrorFormRealm; *cptr && sptr < zptr; *sptr++ = *cptr++);
         sptr += UrlEncodeString (rqptr->AuthRealmPtr, sptr, zptr-sptr);

         if (rqptr->AuthDigestChallengePtr != NULL)
            if (rqptr->AuthDigestChallengePtr[0])
            {
               for (cptr = ErrorFormDigest;
                    *cptr && sptr < zptr;
                    *sptr++ = *cptr++);
               cptr = rqptr->AuthDigestChallengePtr;
               sptr += UrlEncodeString (cptr, sptr, zptr-sptr);
            }

         if (rqptr->AuthBasicChallengePtr != NULL)
            if (rqptr->AuthBasicChallengePtr[0])
            {
               for (cptr = ErrorFormBasic;
                    *cptr && sptr < zptr;
                    *sptr++ = *cptr++);
               cptr = rqptr->AuthBasicChallengePtr;
               sptr += UrlEncodeString (cptr, sptr, zptr-sptr);
            }
      }

      *sptr++ = '\0';
      rqptr->ErrorReportLength = Length = sptr - Buffer;
      rqptr->ErrorReportPtr = VmGetHeap (rqptr, Length);
      memcpy (rqptr->ErrorReportPtr, Buffer, Length);

      return;
   }

   /**********************************/
   /* generate internal error report */
   /**********************************/

   vecptr = FaoVector;
   *vecptr++ = ErrorSourceInfo (SourceModuleName, SourceLineNumber);
   *vecptr++ = cptr = MsgFor(rqptr,MSG_STATUS_ERROR);
   *vecptr++ = rqptr->ResponseStatusCode;
   *vecptr++ = cptr;
   if (rqptr->ProxyTaskPtr != NULL)
      *vecptr++ = MsgFor(rqptr,MSG_STATUS_REPORTED_BY_PROXY);
   else
      *vecptr++ = MsgFor(rqptr,MSG_STATUS_REPORTED_BY_SERVER);
   *vecptr++ = ExplanationPtr;
   *vecptr++ = MsgFor(rqptr,MSG_STATUS_INFO);

   status = sys$faol (&ErrorFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
      ErrorExitVmsStatus (status, ErrorServerFao, FI_LI);

   Buffer[rqptr->ErrorReportLength = Length] = '\0';
   rqptr->ErrorReportPtr = VmGetHeap (rqptr, Length+1);
   memcpy (rqptr->ErrorReportPtr, Buffer, Length);
   rqptr->ErrorReportPtr[rqptr->ErrorReportLength = Length] = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", rqptr->ErrorReportPtr);
}

/*****************************************************************************/
/*
Report general string overflow.
*/

ErrorGeneralOverflow
(
struct RequestStruct *rqptr,
char *SourceModuleName,
int SourceLineNumber
)
{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ErrorGeneralOverflow()\n");

   if (!rqptr->ResponseStatusCode) rqptr->ResponseStatusCode = 500;
   ErrorGeneral (rqptr, MsgFor(rqptr,MSG_GENERAL_OVERFLOW),
                 SourceModuleName, SourceLineNumber);
}

/*****************************************************************************/
/*
*/
 
ErrorInternal
(
struct RequestStruct *rqptr,
int StatusValue,
char *Explanation,
char *SourceModuleName,
int SourceLineNumber
)
{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ErrorInternal()\n");

   rqptr->ResponseStatusCode = 500;

   if (Explanation == NULL)
      ErrorVmsStatus (rqptr, StatusValue, SourceModuleName, SourceLineNumber);
   else
      ErrorGeneral (rqptr, Explanation, SourceModuleName, SourceLineNumber);
}

/*****************************************************************************/
/*
If the server is configured for it return a pointer to some META information
containing softwareID, source code module and line in which the error occured,
otherwise pointer to an empty string.  This function is not used reentrantly
and so provided the contents of the static buffer are used before it is
recalled it will continue to work. The source file format provided by the
"__FILE__" macro will be "device:[directory]name.type;ver".  Reduce that to
"name".  The "__LINE__" macro just provides an integer.
*/
 
char* ErrorSourceInfo
(
char *SourceModuleName,
int SourceLineNumber
) 
{
   static $DESCRIPTOR (MetaSourceInfoFaoDsc,
"<META NAME=\"generator\" CONTENT=\"!AZ\">\n\
<META NAME=\"module\" CONTENT=\"!AZ\">\n\
<META NAME=\"line\" CONTENT=\"!UL\">\n\0");

   static char  SourceInfo [512];
   static $DESCRIPTOR (SourceInfoDsc, SourceInfo);

   int  status;
 
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ErrorSourceInfo()\n");

   if (!Config.ErrorSourceInfo) return ("");

   status = sys$fao (&MetaSourceInfoFaoDsc, 0, &SourceInfoDsc,
                     SoftwareID, SourceModuleName, SourceLineNumber);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
      ErrorExitVmsStatus (status, ErrorServerFao, FI_LI);

   return (SourceInfo);
}
 
/*****************************************************************************/
/*
*/
 
char* SysGetMsg (int StatusValue)
 
{
   static char  Message [256];
   short int  Length;
   $DESCRIPTOR (MessageDsc, Message);
 
   sys$getmsg (StatusValue, &Length, &MessageDsc, 0, 0);
   Message[Length] = '\0';
   if (Debug) fprintf (stdout, "SysGetMsg() |%s|\n", Message);
   return (Message);
}
 
/****************************************************************************/

