/*****************************************************************************/
/*
                               ProxyMaint.c

See PROXYCACHE.C for commentary.  This module implements both the ROUTINE and
REACTIVE purge maintenance for the proxy file cache.  See PROXYCACHE.C module
for information on cache management strategy.


VERSION HISTORY
---------------
22-OCT-1999  MGD  inaccessable cache device non-fatal during startup
20-JUN-1999  MGD  allow for cache devices >9GB in space calculations,
                  some refinement to statistics report
19-AUG-1998  MGD  initial development (recommenced DEC 1998)
*/
/*****************************************************************************/

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>

/* VMS related header files */
#include <atrdef.h>
#include <descrip.h>
#include <dvidef.h>
#include <fibdef.h>
#include <iodef.h>
#include <rmsdef.h>
#include <ssdef.h>
#include <stsdef.h>

/* application-related header files */
#include "wasd.h"

#define WASD_MODULE "PROXYMAINT"

/******************/
/* global storage */
/******************/

char ErrorProxyMaintTooManyDevices [] = "Volume set has too many members.";

/* i.e. [000000...]*.HTC;* */
char  ProxyMaintScanFileSpec [] = PROXY_CACHE_SPEC;
int ProxyMaintScanFileSpecLength = sizeof(ProxyMaintScanFileSpec)-1;

int  ProxyMaintAllocBlocks,
     ProxyMaintDeletedAllocBlocks,
     ProxyMaintDeletedCount,
     ProxyMaintDeletedUsedBlocks,
     ProxyMaintFileAllocBlocks,
     ProxyMaintFileCount,
     ProxyMaintFileUsedBlocks,
     ProxyMaintPurgeAtHour,
     ProxyMaintPurgeHoursIndex,
     ProxyMaintReactivePurgeCount,
     ProxyMaintResultFileNameLength,
     ProxyMaintRoutinePurgeCount,
     ProxyMaintRoutineHoursIndex,
     ProxyMaintScanType,
     ProxyMaintStatScanCount,
     ProxyMaintTargetPercent;
     ProxyMaintUsedBlocks;

char  ProxyMaintEraseFileName [256],
      ProxyMaintExpandedFileName [256],
      ProxyMaintStatusStringReactive [128] = "0",
      ProxyMaintStatusStringRoutine [128] = "0",
      ProxyMaintStatusStringStatScan [128] = "0",
      ProxyMaintResultFileName [256];

struct FAB  ProxyMaintFab;
struct NAM  ProxyMaintNam;
struct FAB  ProxyMaintEraseFab;
struct NAM  ProxyMaintEraseNam;

struct  ProxyCacheFileAcpStruct  ProxyMaintFileAcpData;

/********************/
/* external storage */
/********************/

#ifdef DBUG
extern boolean Debug;
#else
#define Debug 0 
#endif

extern boolean  AdminWatchEnabled,
                MonitorEnabled,
                ProxyCacheEnabled,
                ProxyCacheFreeSpaceAvailable,
                ProxyAddForwardedByEnabled,
                ProxyReportLog,
                ProxyReportCacheLog,
                ProxyServingEnabled;

extern int  ProxyCacheDeviceCheckMinutes,
            ProxyCacheDeviceMaxPercent,
            ProxyCacheDevicePurgePercent,
            ProxyCacheDeviceTargetPercent,
            ProxyCacheFileKBytesMax,
            ProxyCachePurgeList[],
            ProxyCachePurgeListCount,
            ProxyCacheReloadList[],
            ProxyCacheRoutineHourOfDay,
            ProxyHostCachePurgeHours,
            ProxyHostLookupRetryCount;

extern char  ErrorSanityCheck[],
             ServerHostPort[],
             Utility[];

extern char  *ProxyCacheReloadListPtr,
             *ProxyCachePurgeListPtr;

extern struct AccountingStruct  Accounting;
extern struct ConfigStruct  Config;
extern struct MsgStruct  Msgs;
extern struct ProxyAccountingStruct  ProxyAccounting;

/****************************************************************************/
/*
*/

ProxyMaintInit ()

{
   int  status;
   unsigned short  Channel;
   $DESCRIPTOR (DevNameDsc, "");

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ProxyMaintInit()\n");

   /* check the proxy cache device exists */
   DevNameDsc.dsc$w_length = strlen(PROXY_CACHE_ROOT);
   DevNameDsc.dsc$a_pointer = PROXY_CACHE_ROOT;
   if (VMSnok (status = sys$assign (&DevNameDsc, &Channel, 0, 0)))
   {
      if (status == SS$_IVDEVNAM)
      {
         fprintf (stdout,
"%%%s-W-PROXYMAINT, cache device problem - PROXY CACHING DISABLED\n",
                  Utility);
         ProxyCacheEnabled = false;
         return;
      }
      ErrorExitVmsStatus (status, PROXY_CACHE_ROOT "000000]", FI_LI);
   }
   sys$dassgn (Channel);

   /* check if the absolute maximum is being exceeded */
   ProxyCacheFreeSpaceAvailable = ProxyAccounting.FreeSpaceAvailable =
      ProxyMaintDeviceCheckUsed (ProxyCacheDeviceMaxPercent);

   /* if space available schedule a purge, else if exceeded begin a purge */
   if (ProxyCacheFreeSpaceAvailable)
   {
      ProxyMaintPrintDeviceStats ();
      ProxyMaintReactiveTimer (0);
   }
   else
      ProxyMaintScanBegin (PROXY_MAINT_SCAN_REACTIVE);

   /* schedule a routine purge */
   ProxyMaintRoutineTimer (0);
}

/****************************************************************************/
/*
Print cache device statistics to stdout.
*/

ProxyMaintPrintDeviceStats ()

{
   int  FreeBlocks,
        FreeMBytes,
        FreePercent,
        TotalMBytes,
        TotalBlocks,
        UsedBlocks,
        UsedMBytes,
        UsedPercent;
   char  *DevNamePtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ProxyMaintPrintDeviceStats()\n");

   ProxyMaintDeviceStats (&DevNamePtr, &TotalBlocks, &UsedBlocks, &FreeBlocks);
   TotalMBytes = PROXY_MAINT_DEVICE_MBYTES(TotalBlocks);
   UsedMBytes = PROXY_MAINT_DEVICE_MBYTES(UsedBlocks);
   FreeMBytes = PROXY_MAINT_DEVICE_MBYTES(FreeBlocks);
   UsedPercent = PROXY_MAINT_DEVICE_PERCENT_USED(TotalBlocks,FreeBlocks);
   FreePercent = PROXY_MAINT_DEVICE_PERCENT_FREE(TotalBlocks,FreeBlocks);

   fprintf (stdout,
"%%%s-I-PROXYMAINT, cache device %s\n\
%d blocks (%dMB), %d used (%dMB %d%%), %d free (%dMB %d%%)\n",
            Utility, DevNamePtr,
            TotalBlocks, TotalMBytes,
            UsedBlocks, UsedMBytes, UsedPercent,
            FreeBlocks, FreeMBytes, FreePercent);
}

/****************************************************************************/
/*
Include cache device statistics in WATCH report.
*/

ProxyMaintWatchDeviceStats ()

{
   int  FreeBlocks,
        FreeMBytes,
        FreePercent,
        TotalMBytes,
        TotalBlocks,
        UsedBlocks,
        UsedMBytes,
        UsedPercent;
   char  *DevNamePtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ProxyMaintWatchDeviceStats()\n");

   ProxyMaintDeviceStats (&DevNamePtr, &TotalBlocks, &UsedBlocks, &FreeBlocks);
   TotalMBytes = PROXY_MAINT_DEVICE_MBYTES(TotalBlocks);
   UsedMBytes = PROXY_MAINT_DEVICE_MBYTES(UsedBlocks);
   FreeMBytes = PROXY_MAINT_DEVICE_MBYTES(FreeBlocks);
   UsedPercent = PROXY_MAINT_DEVICE_PERCENT_USED(TotalBlocks,FreeBlocks);
   FreePercent = PROXY_MAINT_DEVICE_PERCENT_FREE(TotalBlocks,FreeBlocks);

   AdminWatchDataFormatted (
"!AZ !UL blocks (!ULMB), !UL used (!ULMB !UL%), !UL free (!ULMB !UL%)\n", 
      DevNamePtr, TotalBlocks, TotalMBytes,
      UsedBlocks, UsedMBytes, UsedPercent,
      FreeBlocks, FreeMBytes, FreePercent);
}

/****************************************************************************/
/*
Return information about file system space on the proxy cache device.  Will
function correctly with volume sets of up to eight members.
*/

ProxyMaintDeviceStats
(
char **DevNamePtrPtr,
int *TotalBlocksPtr,
int *UsedBlocksPtr,
int *FreeBlocksPtr
)
{
#define PROXY_MAINT_CACHE_DEVICE_MAX 8

   static int  DeviceCount,
               FreeBlocks,
               MaxBlock;
   static unsigned short  Length;
   static short  DevChannel [PROXY_MAINT_CACHE_DEVICE_MAX];
   static char  CacheDevName [65],
                DevName [65];
   static $DESCRIPTOR (DevNameDsc, "");
   static struct {
      short BufferLength;
      short ItemCode;
      void  *BufferPtr;
      void  *LengthPtr;
   }
   DevNamItemList [] = 
   {
      { sizeof(CacheDevName), DVI$_DEVNAM, &CacheDevName, &Length },
      { 0, 0, 0, 0 }
   },
   NextDevNamItemList [] = 
   {
      { sizeof(DevName), DVI$_NEXTDEVNAM, &DevName, &Length },
      { 0, 0, 0, 0 }
   },
   BlocksItemList [] = 
   {
      { sizeof(MaxBlock), DVI$_MAXBLOCK, &MaxBlock, 0 },
      { sizeof(FreeBlocks), DVI$_FREEBLOCKS, &FreeBlocks, 0 },
      { 0, 0, 0, 0 }
   };

   int  idx,
        status,
        TotalBlocks,
        TotalFreeBlocks,
        TotalUsedBlocks;
   struct AnIOsb  IOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ProxyMaintDeviceStats()\n");

   if (!DeviceCount)
   {
      /**************/
      /* initialize */
      /**************/

      DevNameDsc.dsc$w_length = strlen(PROXY_CACHE_ROOT);
      DevNameDsc.dsc$a_pointer = PROXY_CACHE_ROOT;

      /* assign a channel to the cache device (or primary if a volument set) */
      if (VMSnok (status =
          sys$assign (&DevNameDsc, &DevChannel[DeviceCount++], 0, 0)))
         ErrorExitVmsStatus (status, PROXY_CACHE_ROOT "000000]", FI_LI);

      status = sys$getdvi (0, DevChannel[0], 0,
                           &DevNamItemList, &IOsb, 0, 0, 0);
      if (Debug)
         fprintf (stdout, "sys$getdvi() %%X%08.08X IOsb: %%X%08.08X\n",
                  status, IOsb.Status);

      if (VMSok (status)) status = IOsb.Status;
      if (VMSnok (status))
         ErrorExitVmsStatus (status, "sys$getdvi()", FI_LI);

      CacheDevName[Length] = '\0';
      if (CacheDevName[0] == '_')
         memmove (CacheDevName, CacheDevName+1, Length);
      if (Debug) fprintf (stdout, "|%s|\n", CacheDevName);

      /* loop assigning a channel to all devices in volume set (if it is!) */
      for (;;)
      {
         if (DeviceCount >= PROXY_MAINT_CACHE_DEVICE_MAX)
            ErrorExitVmsStatus (0, ErrorProxyMaintTooManyDevices, FI_LI);

         status = sys$getdvi (0, DevChannel[0], 0,
                              &NextDevNamItemList, &IOsb, 0, 0, 0);
         if (Debug)
            fprintf (stdout, "sys$getdvi() %%X%08.08X IOsb: %%X%08.08X\n",
                     status, IOsb.Status);

         if (VMSok (status)) status = IOsb.Status;
         if (VMSnok (status))
            ErrorExitVmsStatus (status, "sys$getdvi()", FI_LI);

         DevName[Length] = '\0';
         if (Debug) fprintf (stdout, "|%s|\n", DevName);

         if (!Length) break;

         DevNameDsc.dsc$w_length = Length;
         DevNameDsc.dsc$a_pointer = DevName;

         if (VMSnok (status =
             sys$assign (&DevNameDsc, &DevChannel[DeviceCount++], 0, 0)))
            ErrorExitVmsStatus (status, DevName, FI_LI);
      }

      if (Debug) fprintf (stdout, "DeviceCount: %d\n", DeviceCount);
   }

   /***********/
   /* process */
   /***********/

   TotalBlocks = TotalFreeBlocks = 0;

   for (idx = 0; idx < DeviceCount; idx++)
   {
      status = sys$getdvi (0, DevChannel[idx], 0,
                           &BlocksItemList, &IOsb, 0, 0, 0);
      if (Debug)
         fprintf (stdout, "sys$getdvi() %%X%08.08X IOsb: %%X%08.08X\n",
                  status, IOsb.Status);
      if (VMSok (status)) status = IOsb.Status;
      if (VMSnok (status))
         ErrorExitVmsStatus (status, "sys$getdvi()", FI_LI);

      if (Debug) fprintf (stdout, "%d %d\n", MaxBlock, FreeBlocks);

      TotalBlocks += MaxBlock;
      TotalFreeBlocks += FreeBlocks;
   }

   TotalUsedBlocks = TotalBlocks - TotalFreeBlocks;

   if (Debug)
      fprintf (stdout, "%d %d %d %dMB %dMB %dMB %d%% %d%%\n",
               TotalBlocks, TotalFreeBlocks, TotalUsedBlocks,
               TotalBlocks >> 11, TotalFreeBlocks >> 11, TotalUsedBlocks >> 11,
               TotalFreeBlocks*100/TotalBlocks, 
               TotalUsedBlocks*100/TotalBlocks);

   if (DevNamePtrPtr != NULL) *DevNamePtrPtr = CacheDevName;
   if (TotalBlocksPtr != NULL) *TotalBlocksPtr = TotalBlocks;
   if (UsedBlocksPtr != NULL) *UsedBlocksPtr = TotalUsedBlocks;
   if (FreeBlocksPtr != NULL) *FreeBlocksPtr = TotalFreeBlocks;
}

/****************************************************************************/
/*
If the percentage of used space on the cache device is greater than or equal to
the specified percentage then return true, otherwise return false.
*/

boolean ProxyMaintDeviceCheckUsed (int LimitPercent)

{
   int  status,
        FreeBlocks,
        UsedPercent,
        TotalBlocks;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "ProxyMaintDeviceCheckUsed() %d\n", LimitPercent);

   ProxyMaintDeviceStats (NULL, &TotalBlocks, NULL, &FreeBlocks);
   UsedPercent = PROXY_MAINT_DEVICE_PERCENT_USED(TotalBlocks,FreeBlocks);
   if (Debug) fprintf (stdout, "UsedPercent: %d\n", UsedPercent);

   if (UsedPercent > LimitPercent)
      return (false);
   else
      return (true);
}

/*****************************************************************************/
/*
Schedules a timer event for every 'configuration parameter' minutes.  When
timer AST is delivered to this function the cache device free space is checked
against the 'configuration parameter' maximum percentage.  If less a purge
cycle is initiated.  If acceptable another timer event is queued.  This
function is also called from ProxyMaintInit() to initiate, and from
ProxyMaintScanEnd() to schedule another free space check.
*/

ProxyMaintReactiveTimer (int RequestId)

{
   int  status;
   unsigned long  PurgeDelta [2];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ProxyMaintReactiveTimer() %d\n", RequestId);

   /* in case there's an outstanding timer */
   status = sys$cantim (&ProxyMaintReactiveTimer, 0);
   if (Debug) fprintf (stdout, "sys$cantim() %%X%08.08X\n", status);

   if (RequestId)
   {
      ProxyCacheFreeSpaceAvailable = ProxyAccounting.FreeSpaceAvailable =
         ProxyMaintDeviceCheckUsed (ProxyCacheDeviceMaxPercent);

      /* insufficient free space, there used to be, no other purge underway */
      if (!ProxyCacheFreeSpaceAvailable &&
          ProxyMaintScanType == PROXY_MAINT_SCAN_NONE)
      {
         /*********/
         /* purge */
         /*********/

         ProxyMaintScanBegin (PROXY_MAINT_SCAN_REACTIVE);
         return;
      }

      if (AdminWatchEnabled & ADMIN_WATCH_PROXY_CACHE_MNT)
      {
         if (ProxyCacheFreeSpaceAvailable)
            AdminWatchThis (NULL, FI_LI, ADMIN_WATCH_PROXY_CACHE_MNT,
                            "MAINT sufficient free SPACE !UL% max",
                            ProxyCacheDeviceMaxPercent);
         else
            AdminWatchThis (NULL, FI_LI, ADMIN_WATCH_PROXY_CACHE_MNT,
                            "MAINT INSUFFICIENT free SPACE !UL% max",
                            ProxyCacheDeviceMaxPercent);
         ProxyMaintWatchDeviceStats ();
      }

      /* drop though to set a timer for the next check */
   }

   /* this is one minute delta */
   PurgeDelta[0] = -600000000;
   PurgeDelta[1] = -1;
   /* multiply by the number of minutes between free space checks */
   if (VMSnok (status =
       lib$mult_delta_time (&ProxyCacheDeviceCheckMinutes, &PurgeDelta)))
      ErrorExitVmsStatus (status, "lib$mult_delta_time()", FI_LI);

   /* queue up a timer event scheduling the next free space check */
   if (VMSnok (status =
       sys$setimr (0, &PurgeDelta, &ProxyMaintReactiveTimer,
                   &ProxyMaintReactiveTimer, 0)))
      ErrorExitVmsStatus (status, "sys$setimr()", FI_LI);
}

/*****************************************************************************/
/*
Schedule a routine purge for the configuration hour of today (if before that
hour) or tomorrow.
*/

ProxyMaintRoutineTimer (boolean RequestId)

{
   static unsigned short  NumTime [7];
   static unsigned long  BinTime [2],
                         DeltaTime [2],
                         NextTime [2];
   static char  String [32];
   static $DESCRIPTOR (StringDsc, String);
   static $DESCRIPTOR (TodayFaoDsc, "0 !UL:00:00.00");
   static $DESCRIPTOR (TomorrowFaoDsc, "1 !UL:00:00.00");

   int  status;
   unsigned short  Length;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ProxyMaintRoutineTimer() %d\n", RequestId);

   /* in case there's an outstanding timer */
   status = sys$cantim (&ProxyMaintRoutineTimer, 0);
   if (Debug) fprintf (stdout, "sys$cantim() %%X%08.08X\n", status);

   /* no other purge is underway, start a routine one */
   if (RequestId && ProxyMaintScanType == PROXY_MAINT_SCAN_NONE)
   {
      /*********/
      /* purge */
      /*********/

      ProxyMaintScanBegin (PROXY_MAINT_SCAN_ROUTINE);
   }

   sys$gettim (&BinTime);
   sys$numtim (&NumTime, &BinTime);

   /* either today (if soon enough) or tomorrow */
   StringDsc.dsc$w_length = sizeof(String)-1;
   if (NumTime[3] < ProxyCacheRoutineHourOfDay)
      status = sys$fao (&TodayFaoDsc, &Length, &StringDsc,
                        ProxyCacheRoutineHourOfDay);
   else
      status = sys$fao (&TomorrowFaoDsc, &Length, &StringDsc,
                        ProxyCacheRoutineHourOfDay);
   if (Debug) fprintf (stdout, "sys$fao() %%X%08.08X\n", status);
   String[StringDsc.dsc$w_length = Length] = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", String);
      
   /* back to midnight this same day */
   NumTime[3] = NumTime[4] = NumTime[5] = NumTime[6] = 0;
   status = lib$cvt_vectim (&NumTime, &DeltaTime);
   if (Debug) fprintf (stdout, "lib$cvt_vectim() %%X%08.08X\n", status);

   /* add the midnight and the generated today/tomorrow delta time */
   status = sys$bintim (&StringDsc, &BinTime);
   if (Debug) fprintf (stdout, "sys$bintim() %%X%08.08X\n", status);
   status = lib$add_times(&BinTime, &DeltaTime, &NextTime);
   if (Debug) fprintf (stdout, "lib$add_times() %%X%08.08X\n", status);

   if (ProxyReportLog)
   {
      /* for the status message convert it to ASCII */
      StringDsc.dsc$w_length = sizeof(String)-1;
      status = sys$asctim (&Length, &StringDsc, &NextTime, 0);
      if (Debug) fprintf (stdout, "sys$asctim() %%X%08.08X\n", status);
      String[17] = '\0';
      if (String[0] == ' ') String[0] = '0';
      fprintf (stdout, "%%%s-I-PROXYMAINT, next routine purge begins %s\n",
               Utility, String);
   }

   if (VMSnok (status =
       sys$setimr (0, &NextTime, &ProxyMaintRoutineTimer,
                   &ProxyMaintRoutineTimer, 0)))
      ErrorExitVmsStatus (status, "sys$setimr()", FI_LI);
}

/****************************************************************************/
/*
Initialize ready to begin a routine or purge pass.
*/

ProxyMaintScanBegin (int ScanType)

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ProxyMaintScanBegin()\n");

   /* just extra careful! */
   if (ProxyMaintScanType != PROXY_MAINT_SCAN_NONE) return (false);

   ProxyMaintScanType = ScanType;
   ProxyMaintTargetPercent = ProxyCacheDeviceTargetPercent;
   ProxyMaintPurgeAtHour = ProxyCachePurgeList[ProxyMaintPurgeHoursIndex = 0];

   ProxyMaintDeletedCount =
      ProxyMaintDeletedAllocBlocks =
      ProxyMaintDeletedUsedBlocks =
      ProxyMaintAllocBlocks =
      ProxyMaintUsedBlocks = 0;

   switch (ProxyMaintScanType)
   {
      case PROXY_MAINT_SCAN_STATISTICS :

         /****************************/
         /* getting cache statistics */
         /****************************/

         if (AdminWatchEnabled & ADMIN_WATCH_PROXY_CACHE_MNT)
         {
            AdminWatchThis (NULL, FI_LI, ADMIN_WATCH_PROXY_CACHE_MNT,
               "MAINT begin STATISTICS SCAN");
            ProxyMaintWatchDeviceStats ();
         }

         if (ProxyReportLog)
         {
            FaoPrintStdout ("%!AZ-I-PROXYMAINT, !AZ, begin statistics scan\n",
                            Utility, DateTime(NULL, 20));
            ProxyMaintPrintDeviceStats ();
         }

         ProxyMaintStatScanCount++;

         FaoPrint (ProxyMaintStatusStringStatScan,
                   sizeof(ProxyMaintStatusStringStatScan),
                   NULL, "!UL, !20%D<BR><I>in-progress</I>",
                   ProxyMaintStatScanCount, 0);

         FaoPrint (ProxyAccounting.StatusString,
                   sizeof(ProxyAccounting.StatusString),
                   NULL, "!AZ, STATSCAN in-progress", DigitDayTime(0));

         ProxyMaintScanNow ();
         return (true);

      case PROXY_MAINT_SCAN_REACTIVE :

         /********************************/
         /* purging to create free space */
         /********************************/

         if (AdminWatchEnabled & ADMIN_WATCH_PROXY_CACHE_MNT)
         {
            AdminWatchThis (NULL, FI_LI, ADMIN_WATCH_PROXY_CACHE_MNT,
               "MAINT begin REACTIVE PURGE >!UL hours !UL% max !UL% target",
               ProxyMaintPurgeAtHour, ProxyCacheDeviceMaxPercent,
               ProxyMaintTargetPercent);
            ProxyMaintWatchDeviceStats ();
         }

         if (ProxyReportLog)
         {
            FaoPrintStdout (
"%!AZ-I-PROXYMAINT, !AZ, begin reactive purge\n\
(>!UL hours, !UL% max, !UL% target)\n",
                      Utility, DateTime(NULL, 20),
                      ProxyMaintPurgeAtHour,  ProxyCacheDeviceMaxPercent,
                      ProxyMaintTargetPercent);
            ProxyMaintPrintDeviceStats ();
         }

         ProxyMaintReactivePurgeCount++;

         FaoPrint (ProxyMaintStatusStringReactive,
                   sizeof(ProxyMaintStatusStringReactive),
                   NULL, "!UL, !20%D<BR><I>in-progress</I>",
                   ProxyMaintReactivePurgeCount, 0);

         FaoPrint (ProxyAccounting.StatusString,
                   sizeof(ProxyAccounting.StatusString), NULL,
"!AZ, REACTIVE purge, >!UL hours, !UL% max, !UL% target",
                   DigitDayTime(0), ProxyMaintPurgeAtHour,
                   ProxyCacheDeviceMaxPercent, ProxyMaintTargetPercent);

         ProxyMaintScanNow ();
         return (true);

      case PROXY_MAINT_SCAN_ROUTINE :

         /*******************/
         /* routine cleanup */
         /*******************/

         if (AdminWatchEnabled & ADMIN_WATCH_PROXY_CACHE_MNT)
         {
            AdminWatchThis (NULL, FI_LI, ADMIN_WATCH_PROXY_CACHE_MNT,
               "MAINT begin ROUTINE PURGE >!UL hours",
               ProxyMaintPurgeAtHour);
            ProxyMaintWatchDeviceStats ();
         }

         if (ProxyReportLog)
         {
            FaoPrintStdout (
"%!AZ-I-PROXYMAINT, !AZ, begin routine purge (>!UL hours)\n",
                      Utility, DateTime(NULL, 20), ProxyMaintPurgeAtHour);
            ProxyMaintPrintDeviceStats ();
         }

         ProxyMaintRoutinePurgeCount++;

         FaoPrint (ProxyMaintStatusStringRoutine,
                   sizeof(ProxyMaintStatusStringRoutine),
                   NULL, "!UL, !20%D<BR><I>in-progress</I>",
                   ProxyMaintRoutinePurgeCount, 0);

         FaoPrint (ProxyAccounting.StatusString,
                   sizeof(ProxyAccounting.StatusString),
                   NULL, "!AZ, ROUTINE in-progress", DigitDayTime(0));

         ProxyMaintScanNow ();
         return (true);

      default :
         ErrorExitVmsStatus (SS$_BUGCHECK, ErrorSanityCheck, FI_LI);
   }

   return (false);
}

/****************************************************************************/
/*
End of a single pass through the entire cache directory tree.  A PURGE will be
made up of multiple passes, each using the next access hours found in the
'ProxyCachePurgeList' array.  A ROUTINE purge will only make the one pass
though the tree using only the hours in the zero element.
*/

ProxyMaintScanEnd ()

{
   int  status,
        FreeBlocks,
        FreePercent,
        TotalBlocks;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ProxyMaintScanEnd()\n");

   ProxyCacheFreeSpaceAvailable = ProxyAccounting.FreeSpaceAvailable =
      ProxyMaintDeviceCheckUsed (ProxyMaintTargetPercent);

   switch (ProxyMaintScanType)
   {
      case PROXY_MAINT_SCAN_STATISTICS :

         /*******************/
         /* cach statistics */
         /*******************/

         if (AdminWatchEnabled & ADMIN_WATCH_PROXY_CACHE_MNT)
         {
            AdminWatchThis (NULL, FI_LI, ADMIN_WATCH_PROXY_CACHE_MNT,
               "MAINT end STATISTICS SCAN !UL files (!UL/!UL blocks)",
               ProxyMaintFileCount, ProxyMaintUsedBlocks,
               ProxyMaintAllocBlocks);
            ProxyMaintWatchDeviceStats ();
         }

         if (ProxyReportLog)
         {
            FaoPrintStdout (
"%!AZ-I-PROXYMAINT, !AZ, end statistics scan\n!UL files (!UL/!UL)\n",
                      Utility, DateTime(NULL, 20),
                      ProxyMaintFileCount, ProxyMaintUsedBlocks,
                      ProxyMaintAllocBlocks);
            ProxyMaintPrintDeviceStats ();
         }

         FaoPrint (ProxyMaintStatusStringStatScan,
                   sizeof(ProxyMaintStatusStringStatScan), NULL,
                   "!UL, !20%D<BR>!UL files (!UL/!UL)",
                   ProxyMaintStatScanCount,
                   0, ProxyMaintFileCount, ProxyMaintUsedBlocks,
                   ProxyMaintAllocBlocks);

         FaoPrint (ProxyAccounting.StatusString,
                   sizeof(ProxyAccounting.StatusString), NULL,
                   "!AZ, !UL files (!UL/!UL)",
                   DigitDayTime(0), ProxyMaintFileCount,
                   ProxyMaintUsedBlocks, ProxyMaintAllocBlocks);

         ProxyMaintScanType = PROXY_MAINT_SCAN_NONE;
         break;

      case PROXY_MAINT_SCAN_REACTIVE :

         /********************************/
         /* purging to create free space */
         /********************************/

         /* if reached desired percent, or one complete purge, or no files */
         if (ProxyCacheFreeSpaceAvailable ||
             ProxyMaintPurgeHoursIndex >= ProxyCachePurgeListCount ||
             !ProxyMaintFileCount)
         {
            /*****************/
            /* cease purging */
            /*****************/

            if (AdminWatchEnabled & ADMIN_WATCH_PROXY_CACHE_MNT)
            {
               AdminWatchThis (NULL, FI_LI, ADMIN_WATCH_PROXY_CACHE_MNT,
"MAINT end REACTIVE PURGE \
!UL files (!UL/!UL blocks) !UL deleted (!UL/!UL blocks)",
                  ProxyMaintFileCount, ProxyMaintUsedBlocks,
                  ProxyMaintAllocBlocks, ProxyMaintDeletedCount,
                  ProxyMaintDeletedUsedBlocks, ProxyMaintDeletedAllocBlocks);
               ProxyMaintWatchDeviceStats ();
            }

            if (ProxyReportLog)
            {
               FaoPrintStdout (
"%!AZ-I-PROXYMAINT, !AZ, end reactive purge\n\
!UL files (!UL/!UL), !UL deleted (!UL/!UL)\n",
                         Utility, DateTime(NULL, 20),
                         ProxyMaintFileCount, ProxyMaintUsedBlocks,
                         ProxyMaintAllocBlocks, ProxyMaintDeletedCount,
                         ProxyMaintDeletedUsedBlocks,
                         ProxyMaintDeletedAllocBlocks);
               ProxyMaintPrintDeviceStats ();
            }

            FaoPrint (ProxyMaintStatusStringReactive,
                      sizeof(ProxyMaintStatusStringReactive), NULL, 
"!UL, !20%D<BR>!UL files (!UL/!UL), !UL deleted (!UL/!UL)",
                      ProxyMaintReactivePurgeCount, 0,
                      ProxyMaintFileCount, ProxyMaintUsedBlocks,
                      ProxyMaintAllocBlocks, ProxyMaintDeletedCount,
                      ProxyMaintDeletedUsedBlocks,
                      ProxyMaintDeletedAllocBlocks);

            FaoPrint (ProxyAccounting.StatusString,
                      sizeof(ProxyAccounting.StatusString), NULL,
"!AZ, !UL files (!UL/!UL), !UL deleted (!UL/!UL)",
                      DigitDayTime(0), ProxyMaintFileCount,
                      ProxyMaintUsedBlocks, ProxyMaintAllocBlocks,
                      ProxyMaintDeletedCount, ProxyMaintDeletedUsedBlocks,
                      ProxyMaintDeletedAllocBlocks);

            /* set the timer for the next purge check */
            ProxyMaintReactiveTimer (false);

            ProxyMaintScanType = PROXY_MAINT_SCAN_NONE;
            break;
         }

         /**********************/
         /* next pass of purge */
         /**********************/

         ProxyMaintPurgeAtHour =
            ProxyCachePurgeList[++ProxyMaintPurgeHoursIndex];

         if (AdminWatchEnabled & ADMIN_WATCH_PROXY_CACHE_MNT)
         {
            AdminWatchThis (NULL, FI_LI, ADMIN_WATCH_PROXY_CACHE_MNT,
"MAINT end pass REACTIVE PURGE \
!UL files (!UL/!UL blocks) !UL deleted (!UL/!UL blocks)",
               ProxyMaintFileCount, ProxyMaintUsedBlocks,
               ProxyMaintAllocBlocks, ProxyMaintDeletedCount,
               ProxyMaintDeletedUsedBlocks, ProxyMaintDeletedAllocBlocks);

            AdminWatchThis (NULL, FI_LI, ADMIN_WATCH_PROXY_CACHE_MNT,
               "MAINT next REACTIVE PURGE >!UL hours !UL% max !UL% target",
               ProxyMaintPurgeAtHour, ProxyCacheDeviceMaxPercent,
               ProxyMaintTargetPercent);

            ProxyMaintWatchDeviceStats ();
         }

         if (ProxyReportLog)
         {
            FaoPrintStdout (
"%!AZ-I-PROXYMAINT, !AZ, end reactive pass\n\
!UL files (!UL/!UL), !UL deleted (!UL/!UL)\n\
next pass >!UL hours !UL% max !UL% target\n",
                      Utility, DateTime(NULL, 20),
                      ProxyMaintFileCount, ProxyMaintUsedBlocks,
                      ProxyMaintAllocBlocks, ProxyMaintDeletedCount,
                      ProxyMaintDeletedUsedBlocks,
                      ProxyMaintDeletedAllocBlocks,
                      ProxyMaintPurgeAtHour,  ProxyCacheDeviceMaxPercent,
                      ProxyMaintTargetPercent);
            ProxyMaintPrintDeviceStats ();
         }

         FaoPrint (ProxyAccounting.StatusString,
                   sizeof(ProxyAccounting.StatusString), NULL,
"!AZ, REACTIVE purge, >!UL hours, !UL% max, !UL% target",
                   DigitDayTime(0), ProxyMaintPurgeAtHour,
                   ProxyCacheDeviceMaxPercent, ProxyMaintTargetPercent);

         ProxyMaintScanNow ();
         break;

      case PROXY_MAINT_SCAN_ROUTINE :

         /*******************/
         /* routine cleanup */
         /*******************/

         if (AdminWatchEnabled & ADMIN_WATCH_PROXY_CACHE_MNT)
         {
            AdminWatchThis (NULL, FI_LI, ADMIN_WATCH_PROXY_CACHE_MNT,
"MAINT end ROUTINE PURGE \
!UL files (!UL/!UL blocks) !UL deleted (!UL/!UL blocks)",
               ProxyMaintFileCount, ProxyMaintUsedBlocks,
               ProxyMaintAllocBlocks, ProxyMaintDeletedCount,
               ProxyMaintDeletedUsedBlocks, ProxyMaintDeletedAllocBlocks);
            ProxyMaintWatchDeviceStats ();
         }

         if (ProxyReportLog)
         {
            FaoPrintStdout (
"%!AZ-I-PROXYMAINT, !AZ, end routine purge\n\
!UL files (!UL/!UL), !UL deleted (!UL/!UL)\n",
                      Utility, DateTime(NULL, 20),
                      ProxyMaintFileCount, ProxyMaintUsedBlocks,
                      ProxyMaintAllocBlocks, ProxyMaintDeletedCount,
                      ProxyMaintDeletedUsedBlocks,
                      ProxyMaintDeletedAllocBlocks);
            ProxyMaintPrintDeviceStats ();
         }

         FaoPrint (ProxyMaintStatusStringRoutine,
                   sizeof(ProxyMaintStatusStringRoutine), NULL, 
"!UL, !20%D<BR>!UL files (!UL/!UL), !UL deleted (!UL/!UL)",
                   ProxyMaintRoutinePurgeCount, 0,
                   ProxyMaintFileCount, ProxyMaintUsedBlocks,
                   ProxyMaintAllocBlocks, ProxyMaintDeletedCount,
                   ProxyMaintDeletedUsedBlocks,
                   ProxyMaintDeletedAllocBlocks);

         FaoPrint (ProxyAccounting.StatusString,
                   sizeof(ProxyAccounting.StatusString), NULL,
"!AZ, !UL files (!UL/!UL), !UL deleted (!UL/!UL)",
                   DigitDayTime(0), ProxyMaintFileCount,
                   ProxyMaintUsedBlocks, ProxyMaintAllocBlocks,
                   ProxyMaintDeletedCount, ProxyMaintDeletedUsedBlocks,
                   ProxyMaintDeletedAllocBlocks);

         ProxyMaintScanType = PROXY_MAINT_SCAN_NONE;
         break;

      default :
         ErrorExitVmsStatus (SS$_BUGCHECK, ErrorSanityCheck, FI_LI);
   }

   /* set logical names used by HTTPd monitor utility */
   if (MonitorEnabled) DefineMonitorLogicals (NULL);
}

/****************************************************************************/
/*
Initialize the RMS structures ready for the file search.
*/

ProxyMaintScanNow ()

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "ProxyMaintScanNow() |%s|\n", ProxyMaintScanFileSpec);

   /* set logical names used by HTTPd monitor utility */
   if (MonitorEnabled) DefineMonitorLogicals (NULL);

   ProxyMaintFab = cc$rms_fab;
   ProxyMaintFab.fab$l_fna = ProxyMaintScanFileSpec;
   ProxyMaintFab.fab$b_fns = ProxyMaintScanFileSpecLength;
   ProxyMaintFab.fab$l_nam = &ProxyMaintNam;

   /* initialize the NAM block */
   ProxyMaintNam = cc$rms_nam;
   ProxyMaintNam.nam$l_esa = ProxyMaintExpandedFileName;
   ProxyMaintNam.nam$b_ess = sizeof(ProxyMaintExpandedFileName)-1;
   ProxyMaintNam.nam$l_rsa = ProxyMaintResultFileName;
   ProxyMaintNam.nam$b_rss = sizeof(ProxyMaintResultFileName)-1;

   status = sys$parse (&ProxyMaintFab, 0, 0);
   if (Debug) fprintf (stdout, "sys$parse() %%X%08.08X\n", status);
   if (VMSnok (status))
      ErrorExitVmsStatus (status, "sys$parse()", FI_LI);

   ProxyMaintFileCount = 0;

   /* now we've done the parse do the searching asynchronously */
   ProxyMaintFab.fab$l_fop = FAB$M_ASY;

   ProxyMaintScanSearch ();
}

/****************************************************************************/
/*
Simply queue another asynchronous RMS search.
*/

ProxyMaintScanSearch ()

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ProxyMaintScanSearch()\n");

   switch (ProxyMaintScanType)
   {
      case PROXY_MAINT_SCAN_STATISTICS :

         /* just continue through the complete scan */
         break;

      case PROXY_MAINT_SCAN_REACTIVE :

         /* only recheck free space after every ten files deleted */
         if (ProxyMaintFileCount &&
             (!ProxyMaintDeletedCount || ProxyMaintDeletedCount % 10))
            break;

         if (ProxyMaintDeviceCheckUsed (ProxyMaintTargetPercent))
         {
            /* reached our desired limit, cease purging */
            ProxyMaintScanEnd ();
            return;
         }
         break;

      case PROXY_MAINT_SCAN_ROUTINE :

         /* just continue through the complete routine purge */
         break;

      default :
         ErrorExitVmsStatus (SS$_BUGCHECK, ErrorSanityCheck, FI_LI);
   }

   if (AdminWatchEnabled & ADMIN_WATCH_PROXY_CACHE_MNT)
   {
      if (ProxyMaintFileCount &&
          ProxyMaintFileCount % PROXY_MAINT_SCAN_WATCH_FILES == 0)
      {
         /* WATCH status entry every so many files */
         AdminWatchThis (NULL, FI_LI, ADMIN_WATCH_PROXY_CACHE_MNT,
"MAINT purge STATUS !UL files (!UL/!UL blocks) !UL deleted (!UL/!UL blocks)",
            ProxyMaintFileCount, ProxyMaintUsedBlocks,
            ProxyMaintAllocBlocks, ProxyMaintDeletedCount,
            ProxyMaintDeletedUsedBlocks, ProxyMaintDeletedAllocBlocks);
         ProxyMaintWatchDeviceStats ();
      }
   }

   status = sys$search (&ProxyMaintFab, &ProxyMaintScanSearchAst,
                                        &ProxyMaintScanSearchAst);
   if (Debug) fprintf (stdout, "sys$search() %%X%08.08X\n", status);
}

/****************************************************************************/
/*
AST from sys$search().  Check status and report any error.  Otherwise generate
and queue an asynchronous ACP file information QIO.
*/

ProxyMaintScanSearchAst (struct FAB *FabPtr)

{
   static $DESCRIPTOR (DeviceDsc, "");

   register struct ProxyCacheFileAcpStruct  *faptr;

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "ProxyMaintScanSearchAst() %%X%08.08X\n",
               FabPtr->fab$l_sts);

   if (VMSnok (status = FabPtr->fab$l_sts))
   {
      /****************/
      /* search error */
      /****************/

      if (status == RMS$_NMF || (!ProxyMaintFileCount && status == RMS$_FNF))
      {
         /* end of search */
         ProxyMaintScanEnd ();
         return;
      }
      if (status == RMS$_FNF)
      {
         /* file probably deleted during search, just continue searching */
         ProxyMaintScanSearch ();
         return;
      }

      /* more significant error */
      ErrorExitVmsStatus (status, "sys$search()", FI_LI);
   }

   ProxyMaintFileCount++;

   ProxyMaintResultFileName[ProxyMaintResultFileNameLength =
                            ProxyMaintNam.nam$b_rsl] = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", ProxyMaintResultFileName);

   /************/
   /* file ACP */
   /************/

   faptr = &ProxyMaintFileAcpData;

   /* clean the structure out */
   memset (faptr, sizeof(ProxyMaintFileAcpData), 0);

   /* assign a channel to the disk device containing the file */
   DeviceDsc.dsc$a_pointer = ProxyMaintNam.nam$l_dev;
   DeviceDsc.dsc$w_length = ProxyMaintNam.nam$b_dev;

   status = sys$assign (&DeviceDsc, &faptr->Channel, 0, 0, 0);
   if (Debug) fprintf (stdout, "sys$assign() %%X%08.08X\n", status);
   if (VMSnok (status))
      ErrorExitVmsStatus (status, "sys$assign()", FI_LI);

   /* set up the File Information Block for the ACP interface */
   faptr->FileFibAcpDsc.dsc$w_length = sizeof(struct fibdef);
   faptr->FileFibAcpDsc.dsc$a_pointer = &faptr->FileFib;

   memcpy (&faptr->FileFib.fib$w_did, &ProxyMaintNam.nam$w_did, 6);

   faptr->FileNameAcpDsc.dsc$a_pointer = ProxyMaintNam.nam$l_name;
   faptr->FileNameAcpDsc.dsc$w_length = ProxyMaintNam.nam$b_name +
                                        ProxyMaintNam.nam$b_type +
                                        ProxyMaintNam.nam$b_ver;
   if (Debug)
      fprintf (stdout, "name |%*.*s|\n",
               faptr->FileNameAcpDsc.dsc$w_length,
               faptr->FileNameAcpDsc.dsc$w_length,
               faptr->FileNameAcpDsc.dsc$a_pointer);

   faptr->FileAtr[0].atr$w_size = sizeof(faptr->CdtBinTime);
   faptr->FileAtr[0].atr$w_type = ATR$C_CREDATE;
   faptr->FileAtr[0].atr$l_addr = &faptr->CdtBinTime;
   faptr->FileAtr[1].atr$w_size = sizeof(faptr->RdtBinTime);
   faptr->FileAtr[1].atr$w_type = ATR$C_REVDATE;
   faptr->FileAtr[1].atr$l_addr = &faptr->RdtBinTime;
   faptr->FileAtr[2].atr$w_size = sizeof(faptr->EdtBinTime);
   faptr->FileAtr[2].atr$w_type = ATR$C_EXPDATE;
   faptr->FileAtr[2].atr$l_addr = &faptr->EdtBinTime;
   faptr->FileAtr[3].atr$w_size = sizeof(faptr->AscDates);
   faptr->FileAtr[3].atr$w_type = ATR$C_ASCDATES;
   faptr->FileAtr[3].atr$l_addr = &faptr->AscDates;
   faptr->FileAtr[4].atr$w_size = sizeof(faptr->RecAttr);
   faptr->FileAtr[4].atr$w_type = ATR$C_RECATTR;
   faptr->FileAtr[4].atr$l_addr = &faptr->RecAttr;
   faptr->FileAtr[5].atr$w_size =
      faptr->FileAtr[5].atr$w_type =
      faptr->FileAtr[5].atr$l_addr = 0;

   status = sys$qio (0, faptr->Channel, IO$_ACCESS,
                     &faptr->AcpIOsb, &ProxyMaintAcpInfoAst, 0, 
                     &faptr->FileFibAcpDsc, &faptr->FileNameAcpDsc, 0, 0,
                     &faptr->FileAtr, 0);
   if (Debug) fprintf (stdout, "sys$qio() %%X%08.08X\n", status);
}

/****************************************************************************/
/*
AST from ACP QIO. Calculate the relevant file information and determine if it
should be purged (deleted).  If not then just queue another search.  If it
should be then initialize the deletion RMS structures ready for sys$erase().
*/

ProxyMaintAcpInfoAst ()

{
   register struct ProxyCacheFileAcpStruct  *faptr;

   int  status,
        AccessHours;
   unsigned long  AllocatedVbn,
                  EndOfFileVbn;
   unsigned long  CurrentBinTime[2];

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "ProxyMaintAcpInfoAst() %%X%08.08X\n",
               ProxyMaintFileAcpData.AcpIOsb.Status);

   faptr = &ProxyMaintFileAcpData;

   sys$dassgn (faptr->Channel);

   if (VMSnok (status = faptr->AcpIOsb.Status))
   {
      /*************/
      /* ACP error */
      /*************/

      fprintf (stdout,
         "%%%s-W-PROXYMAINT, file ACP error (%s %d)\n-%s\n \\%s\\\n",
         Utility, FI_LI, SysGetMsg(status)+1, ProxyMaintResultFileName);

      /* file probably deleted during search, just continue searching */
      ProxyMaintScanSearch ();
      return;
   }

   /**************/
   /* ACP QIO OK */
   /**************/

   AllocatedVbn = faptr->RecAttr.AllocatedVbnLo +
                  (faptr->RecAttr.AllocatedVbnHi << 16);
   EndOfFileVbn = faptr->RecAttr.EndOfFileVbnLo +
                  (faptr->RecAttr.EndOfFileVbnHi << 16);
   if (Debug)
      fprintf (stdout,
         "AllocatedVbn: %d EndOfFileVbn: %d FirstFreeByte %d\n",
         AllocatedVbn, EndOfFileVbn, faptr->RecAttr.FirstFreeByte);

   ProxyMaintAllocBlocks += ProxyMaintFileAllocBlocks = AllocatedVbn;
   if (EndOfFileVbn >= 1 && faptr->RecAttr.FirstFreeByte)
      ProxyMaintUsedBlocks += ProxyMaintFileUsedBlocks = EndOfFileVbn;
   else
      ProxyMaintFileUsedBlocks = 0;

   switch (ProxyMaintScanType)
   {
      case PROXY_MAINT_SCAN_STATISTICS :

         /***************/
         /* search next */
         /***************/

         ProxyMaintScanSearch ();
         return;

      case PROXY_MAINT_SCAN_REACTIVE :
      case PROXY_MAINT_SCAN_ROUTINE :

         /* check how long ago the file was last accessed */
         sys$gettim (&CurrentBinTime);
         AccessHours = ProxyCacheAgeHours (&CurrentBinTime, &faptr->EdtBinTime);
         if (Debug) fprintf (stdout, "access: %d\n", AccessHours);

         if (AccessHours < ProxyMaintPurgeAtHour)
         {
            /***************/
            /* search next */
            /***************/

            ProxyMaintScanSearch ();
            return;
         }

         /* drop through to delete the file */
         break;

      default :
         ErrorExitVmsStatus (SS$_BUGCHECK, ErrorSanityCheck, FI_LI);
   }

   /**************/
   /* delete file *
   /**************/

   if (Debug)
      fprintf (stdout, "%d %d\n",
         AccessHours, ProxyCachePurgeList[ProxyMaintPurgeHoursIndex]);

   ProxyMaintEraseFab = cc$rms_fab;
   ProxyMaintEraseFab.fab$l_fna = ProxyMaintResultFileName;
   ProxyMaintEraseFab.fab$b_fns = ProxyMaintResultFileNameLength;
   ProxyMaintEraseFab.fab$l_nam = &ProxyMaintEraseNam;

   /* initialize the NAM block */
   ProxyMaintEraseNam = cc$rms_nam;
   ProxyMaintEraseNam.nam$l_esa = ProxyMaintEraseFileName;
   ProxyMaintEraseNam.nam$b_ess = sizeof(ProxyMaintEraseFileName)-1;

   status = sys$parse (&ProxyMaintEraseFab, &ProxyMaintScanParseAst,
                                            &ProxyMaintScanParseAst);
   if (Debug) fprintf (stdout, "sys$parse() %%X%08.08X\n", status);
}

/****************************************************************************/
/*
AST from sys$parse().  Check status and report any error.  Otherwise perform an
asynchronous earse (delete) on the file.
*/

ProxyMaintScanParseAst (struct FAB *FabPtr)

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "ProxyMaintScanParseAst() %%X%08.08X\n",
               FabPtr->fab$l_sts);

   if (VMSnok (status = ProxyMaintEraseFab.fab$l_sts))
   {
      /* parse error (serious) */
      ErrorExitVmsStatus (status, "sys$parse()", FI_LI);
   }

   status = sys$erase (&ProxyMaintEraseFab, &ProxyMaintScanEraseAst,
                                            &ProxyMaintScanEraseAst);
   if (Debug) fprintf (stdout, "sys$erase() %%X%08.08X\n", status);
}

/****************************************************************************/
/*
AST from sys$parse().  Check status and report any error. Initiate a search for
the next file.
*/

ProxyMaintScanEraseAst (struct FAB *FabPtr)

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "ProxyMaintScanEraseAst() %%X%08.08X\n",
               FabPtr->fab$l_sts);

   if (VMSok (status = ProxyMaintEraseFab.fab$l_sts))
   {
      ProxyMaintDeletedCount++;
      ProxyMaintDeletedAllocBlocks += ProxyMaintFileAllocBlocks;
      ProxyMaintDeletedUsedBlocks += ProxyMaintFileUsedBlocks;
   }
   else
   {
      fprintf (stdout,
         "%%%s-W-PROXYMAINT, file erase error (%s %d)\n-%s\n \\%s\\\n",
         Utility, FI_LI, SysGetMsg(status)+1, ProxyMaintEraseFileName);
   }

   /* search next */
   ProxyMaintScanSearch ();
}

/*****************************************************************************/
/*
Return a report and control menu related proxy serving.
*/ 

ProxyMaintReport
(
struct RequestStruct *rqptr,
void *NextTaskFunction
)
{
   static $DESCRIPTOR (ResponseFaoDsc,
"<HTML>\n\
<HEAD>\n\
!AZ\
<TITLE>HTTPd !AZ ... Proxy Report/Maintenance</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1><NOBR>HTTPd !AZ</NOBR></H1>\n\
<H2>Proxy Report/Maintenance</H2>\n\
!AZ\n");

   static $DESCRIPTOR (CacheFaoDsc,
"<FORM METHOD=POST ACTION=\"!AZ\">\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Cache Management</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right VALIGN=top>Cache Device:</TH><TD>!AZ\n\
<BR>!UL &nbsp;blocks&nbsp; (!ULMB)\n\
<BR>!UL &nbsp;used&nbsp; (!ULMB !UL%)\n\
<BR>!UL &nbsp;free&nbsp; (!ULMB !UL%)\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Free Space:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>In-Progress:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right VALIGN=top>Statistics Scan:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right VALIGN=top>Routine Purge:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right VALIGN=top>Reactive Purge:</TH><TD>!AZ</TD></TR>\n\
<TR><TD COLSPAN=2><NOBR><BR>\n\
&nbsp;<INPUT TYPE=submit NAME=statistics VALUE=\" Statistics Scan \">\n\
&nbsp;<INPUT TYPE=submit NAME=routine VALUE=\" Routine Purge \">\n\
&nbsp;<INPUT TYPE=submit NAME=reactive VALUE=\" Reactive Purge \">\n\
</NOBR></TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
</FORM>\n\
\
<FORM METHOD=POST ACTION=\"!AZ\">\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Host Name Cache</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=3 CELLSPACING=0 BORDER=0>\n\
<TR><TH></TH>\
<TH ALIGN=left><U>Name</U></TH>\
<TH ALIGN=left><U>Address</U></TH>\
<TH ALIGN=right><U>Hit</U></TH></TR>\n");

   static char  EndPage [] =
"<TR><TD COLSPAN=4><BR>\n\
&nbsp;<INPUT TYPE=submit NAME=hostname VALUE=\" Purge \">\n\
</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
</FORM>\n\
\
</BODY>\n\
</HTML>\n";

   register unsigned long  *vecptr;

   int  status,
        FreeBlocks,
        FreeMBytes,
        FreePercent,
        TotalMBytes,
        TotalBlocks,
        UsedBlocks,
        UsedMBytes,
        UsedPercent;
   unsigned short  Length;
   unsigned long  FaoVector [64];
   char  *DevNamePtr;
   char  Buffer [4096],
         LastReactiveDayDateTime [48],
         LastRoutineDayDateTime [48],
         LastStatScanDayDateTime [48];
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ProxyMaintReport()\n");

   ProxyMaintDeviceStats (&DevNamePtr, &TotalBlocks, &UsedBlocks, &FreeBlocks);
   TotalMBytes = PROXY_MAINT_DEVICE_MBYTES(TotalBlocks);
   UsedMBytes = PROXY_MAINT_DEVICE_MBYTES(UsedBlocks);
   FreeMBytes = PROXY_MAINT_DEVICE_MBYTES(FreeBlocks);
   UsedPercent = PROXY_MAINT_DEVICE_PERCENT_USED(TotalBlocks,FreeBlocks);
   FreePercent = PROXY_MAINT_DEVICE_PERCENT_FREE(TotalBlocks,FreeBlocks);

   rqptr->ResponsePreExpired = PRE_EXPIRE_ADMIN;
   HTTP_HEADER_200_HTML (rqptr);

   vecptr = FaoVector;

   *vecptr++ = HtmlMetaInfo (rqptr, NULL);

   *vecptr++ = ServerHostPort;
   *vecptr++ = ServerHostPort;
   *vecptr++ = DayDateTime (&rqptr->BinaryTime, 20);

   status = sys$faol (&ResponseFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, 0, Buffer, Length);

   status = ProxyMaintStatisticsReport (rqptr);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   vecptr = FaoVector;

   *vecptr++ = INTERNAL_CONTROL_PROXY;

   *vecptr++ = DevNamePtr;
   *vecptr++ = TotalBlocks;
   *vecptr++ = TotalMBytes;
   *vecptr++ = UsedBlocks;
   *vecptr++ = UsedMBytes;
   *vecptr++ = UsedPercent;
   *vecptr++ = FreeBlocks;
   *vecptr++ = FreeMBytes;
   *vecptr++ = FreePercent;

   if (ProxyCacheFreeSpaceAvailable)
      *vecptr++ = "AVAILABLE";
   else
      *vecptr++ = "<FONT COLOR=\"ff0000\"><B>NOT AVAILABLE</B></FONT>";

   switch (ProxyMaintScanType)
   {
      case PROXY_MAINT_SCAN_STATISTICS :
         *vecptr++ = "<FONT COLOR=\"ff0000\"><B>STATISTICS SCAN</B></FONT>";
         break;
      case PROXY_MAINT_SCAN_REACTIVE :
         *vecptr++ = "<FONT COLOR=\"ff0000\"><B>REACTIVE PURGE</B></FONT>";
         break;
      case PROXY_MAINT_SCAN_ROUTINE :
         *vecptr++ = "<FONT COLOR=\"ff0000\"><B>ROUTINE PURGE</B></FONT>";
         break;
      default :
         *vecptr++ = "<I>none</I>";
   }

   *vecptr++ = ProxyMaintStatusStringStatScan;
   *vecptr++ = ProxyMaintStatusStringRoutine;
   *vecptr++ = ProxyMaintStatusStringReactive;

   *vecptr++ = INTERNAL_CONTROL_PROXY;

   status = sys$faol (&CacheFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, 0, Buffer, Length);

   ProxyResolveHostCache (rqptr, NULL, 0);

   NetWriteBuffered (rqptr, NextTaskFunction, EndPage, sizeof(EndPage)-1);
}

/*****************************************************************************/
/*
Return a report and control menu related proxy serving.
*/ 

ProxyMaintAdjust
(
struct RequestStruct *rqptr,
void *NextTaskFunction
)
{
   static $DESCRIPTOR (ResponseFaoDsc,
"<HTML>\n\
<HEAD>\n\
!AZ\
<TITLE>HTTPd !AZ ... Proxy Report/Maintenance</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1><NOBR>HTTPd !AZ</NOBR></H1>\n\
<H2>Proxy Report/Maintenance</H2>\n\
!AZ\n\
\
<FORM METHOD=POST ACTION=\"!AZ\">\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Adjust Executing Server</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=3 CELLSPACING=0 BORDER=0>\n\
<TR><TH ALIGN=right>Serving:</TH><TD>\n\
<INPUT TYPE=radio NAME=ProxyServingEnabled VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=ProxyServingEnabled VALUE=disabled!AZ>disabled\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Add &quot;Forwarded: by&quot;:</TH><TD>\n\
<INPUT TYPE=radio NAME=ProxyAddForwardedByEnabled VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=ProxyAddForwardedByEnabled VALUE=disabled!AZ>disabled\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Report To Log:</TH><TD>\n\
<INPUT TYPE=radio NAME=ProxyReportLog VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=ProxyReportLog VALUE=disabled!AZ>disabled\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Report Cache To Log:</TH><TD>\n\
<INPUT TYPE=radio NAME=ProxyReportCacheLog VALUE=enabled!AZ>enabled\n\
<INPUT TYPE=radio NAME=ProxyReportCacheLog VALUE=disabled!AZ>disabled\n\
</TD></TR>\n\
<TR><TH ALIGN=right>Host Name Lookup Retry:</TH><TD>\n\
<INPUT TYPE=text SIZE=3 NAME=ProxyHostLookupRetryCount VALUE=!UL>\n\
<FONT SIZE=-1>count</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Host Name Cache Purge:</TH><TD>\n\
<INPUT TYPE=text SIZE=3 NAME=ProxyHostCachePurgeHours VALUE=!UL>\n\
<FONT SIZE=-1>hours</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Maximum Load Size:</TH><TD>\n\
<INPUT TYPE=text SIZE=4 NAME=ProxyCacheFileKBytesMax VALUE=!UL>\n\
<FONT SIZE=-1>kBytes</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Reload List:</TH><TD>\n\
<INPUT TYPE=text SIZE=30 NAME=ProxyCacheReloadList VALUE=\"!AZ\">\n\
<FONT SIZE=-1>hours</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Routine Purge:</TH><TD>\n\
<INPUT TYPE=text SIZE=3 NAME=ProxyCacheRoutineHourOfDay VALUE=!UL>\n\
<FONT SIZE=-1>hour-of-day (00-23)</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Purge List:</TH><TD>\n\
<INPUT TYPE=text SIZE=30 NAME=ProxyCachePurgeList VALUE=\"!AZ\">\n\
<FONT SIZE=-1>hours</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Device Usage Check:</TH><TD>\n\
<INPUT TYPE=text SIZE=3 NAME=ProxyCacheDeviceCheckMinutes VALUE=!UL>\n\
<FONT SIZE=-1>minutes</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Device Usage Max:</TH><TD>\n\
<INPUT TYPE=text SIZE=3 NAME=ProxyCacheDeviceMaxPercent VALUE=!UL>\n\
<FONT SIZE=-1>percent</FONT></TD></TR>\n\
<TR><TH ALIGN=right>Device Reduce By:</TH><TD>\n\
<INPUT TYPE=text SIZE=3 NAME=ProxyCacheDevicePurgePercent VALUE=!UL>\n\
<FONT SIZE=-1>percent</FONT></TD></TR>\n\
<TR><TD COLSPAN=3><BR>\n\
<INPUT TYPE=submit NAME=makechanges VALUE=\" Make Changes \">\n\
<INPUT TYPE=reset VALUE=\" Reset \">\n\
</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
</FORM>\n\
\
</BODY>\n\
</HTML>\n");

#define REPBOOL(b)\
if (b)\
{\
   *vecptr++ = RadioButtonChecked;\
   *vecptr++ = RadioButtonUnchecked;\
}\
else\
{\
   *vecptr++ = RadioButtonUnchecked;\
   *vecptr++ = RadioButtonChecked;\
}

   static char  RadioButtonChecked [] = " CHECKED",
                RadioButtonUnchecked [] = "";

   register unsigned long  *vecptr;

   int  status;
   unsigned short  Length;
   unsigned long  FaoVector [64];
   char  Buffer [4096];
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ProxyMaintAdjustMenu()\n");

   rqptr->ResponsePreExpired = PRE_EXPIRE_ADMIN;
   HTTP_HEADER_200_HTML (rqptr);

   vecptr = FaoVector;

   *vecptr++ = HtmlMetaInfo (rqptr, NULL);

   *vecptr++ = ServerHostPort;
   *vecptr++ = ServerHostPort;
   *vecptr++ = DayDateTime (&rqptr->BinaryTime, 20);

   *vecptr++ = INTERNAL_CONTROL_PROXY;

   REPBOOL (ProxyServingEnabled)
   REPBOOL (ProxyAddForwardedByEnabled)
   REPBOOL (ProxyReportLog)
   REPBOOL (ProxyReportCacheLog)
   *vecptr++ = ProxyHostLookupRetryCount;
   *vecptr++ = ProxyHostCachePurgeHours;
   *vecptr++ = ProxyCacheFileKBytesMax;
   *vecptr++ = ProxyCacheReloadListPtr;
   *vecptr++ = ProxyCacheRoutineHourOfDay;
   *vecptr++ = ProxyCachePurgeListPtr;
   *vecptr++ = ProxyCacheDeviceCheckMinutes;
   *vecptr++ = ProxyCacheDeviceMaxPercent;
   *vecptr++ = ProxyCacheDevicePurgePercent;

   status = sys$faol (&ResponseFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, NextTaskFunction, Buffer, Length);
}

/*****************************************************************************/
/*
Return a report and control menu related proxy serving.
*/ 

ProxyMaintStatisticsReport (struct RequestStruct *rqptr)

{
   static $DESCRIPTOR (ResponseFaoDsc,
"<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Proxy Serving!AZ</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=0 CELLSPACING=0 BORDER=0>\n\
<TR><TD VALIGN=top>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right><U>Method</U></TH>\
<TR><TH ALIGN=right>GET:</TH><TD ALIGN=right>!UL</TD></TR>\n\
<TR><TH ALIGN=right>HEAD:</TH><TD ALIGN=right>!UL</TD></TR>\n\
<TR><TH ALIGN=right>POST:</TH><TD ALIGN=right>!UL</TD></TR>\n\
<TR><TH ALIGN=right>PUT:</TH><TD ALIGN=right>!UL</TD></TR>\n\
<TR><TH ALIGN=right>CONNECT:</TH><TD ALIGN=right>!UL</TD></TR>\n\
<TR><TH ALIGN=right>Total:</TH><TD ALIGN=right>!UL</TD></TR>\n\
</TABLE>\n\
</TD><TD VALIGN=top>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right><U>Traffic</U></TH></TR>\n\
<TR><TH ALIGN=right>&nbsp;Network&nbsp;Rx:</TH><TD ALIGN=right>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Tx:</TH><TD ALIGN=right>!AZ</TD></TR>\n\
<TR><TH ALIGN=right></TH><TD ALIGN=right>!UL%</TD></TR>\n\
<TR><TH ALIGN=right>Cache&nbsp;Rx:</TH><TD ALIGN=right>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Tx:</TH><TD ALIGN=right>!AZ</TD></TR>\n\
<TR><TH ALIGN=right></TH><TD ALIGN=right>!UL%</TD></TR>\n\
<TR><TH ALIGN=right>Total:</TH><TD ALIGN=right>!AZ</TD></TR>\n\
</TABLE>\n\
</TD><TD VALIGN=top>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right><U>Request&nbsp;Cache</U></TH>\
<TR><TH ALIGN=right>Read:</TH><TD ALIGN=right>!UL/!UL</TD><TD>(!UL%)</TD></TR>\n\
<TR><TH ALIGN=right>Write:</TH><TD ALIGN=right>!UL</TD><TD>(!UL%)</TD></TR>\n\
<TR><TD>&nbsp;</TD></TR>\n\
<TR><TH ALIGN=right>&nbsp;<U>Not&nbsp;Cacheable</U></TH>\
<TR><TH ALIGN=right>Request:</TH><TD ALIGN=right>!UL</TD><TD>(!UL%)</TD></TR>\n\
<TR><TH ALIGN=right>Response:</TH><TD ALIGN=right>!UL</TD><TD>(!UL%)</TD></TR>\n\
<TR><TH ALIGN=right>Rx/Tx:</TH><TD ALIGN=right>!UL%</TD></TR>\n\
</TABLE>\n\
</TD><TD VALIGN=top>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right><U>Network</U></TH></TR>\n\
<TR><TH ALIGN=right>Requested:</TH><TD ALIGN=right>!UL</TD><TD>(!UL%)</TD></TR>\n\
<TR><TD>&nbsp;</TD></TR>\n\
<TR><TH ALIGN=right>&nbsp;<U>Host&nbsp;Lookup</U></TH>\
<TR><TH ALIGN=right>Numeric:</TH><TD ALIGN=right>!UL</TD></TR>\n\
<TR><TH ALIGN=right>DNS:</TH><TD ALIGN=right>!UL</TD></TR>\n\
<TR><TH ALIGN=right>Cache:</TH><TD ALIGN=right>!UL</TD></TR>\n\
<TR><TH ALIGN=right>Error:</TH><TD ALIGN=right>!UL</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n");

   static long  Addx2 = 2;

   register unsigned long  *vecptr;

   int  status,
        PercentBytesCache,
        PercentBytesNetwork,
        PercentBytesNotCacheable,
        PercentCountCacheRead,
        PercentCountCacheWrite,
        PercentCountRequestNotCacheable,
        PercentCountResponseNotCacheable,
        PercentCountNetwork;
   unsigned short  Length;
   unsigned long  FaoVector [64],
                  QuadBytesCache [2],
                  QuadBytesNotCacheable [2],
                  QuadBytesRaw [2],
                  QuadBytesTotal [2];
   float  FloatBytesCache,
          FloatBytesNotCacheable,
          FloatBytesNetwork,
          FloatBytesTotal,
          FloatCountCacheRead,
          FloatCountCacheWrite,
          FloatCountRequestNotCacheable,
          FloatCountResponseNotCacheable,
          FloatCountNetwork,
          FloatCountTotal,
          FloatNetworkCount,
          FloatPercent;
   char  Buffer [4096],
         BytesCacheRx [32],
         BytesCacheTx [32],
         BytesNotCacheableRx [32],
         BytesNotCacheableTx [32],
         BytesRawRx [32],
         BytesRawTx [32],
         BytesTotal [32];
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ProxyMaintStatisticReport()\n");

   /*********/
   /* bytes */
   /*********/

   status = lib$addx (&ProxyAccounting.QuadBytesCacheRx,
                      &ProxyAccounting.QuadBytesCacheTx,
                      &QuadBytesCache, &Addx2);
   if (Debug) fprintf (stdout, "lib$addx() %%X%08.08X\n", status);

   status = lib$addx (&ProxyAccounting.QuadBytesNotCacheableRx,
                      &ProxyAccounting.QuadBytesNotCacheableTx,
                      &QuadBytesNotCacheable, &Addx2);
   if (Debug) fprintf (stdout, "lib$addx() %%X%08.08X\n", status);

   status = lib$addx (&ProxyAccounting.QuadBytesRawRx,
                      &ProxyAccounting.QuadBytesRawTx,
                      &QuadBytesRaw, &Addx2);
   if (Debug) fprintf (stdout, "lib$addx() %%X%08.08X\n", status);

   status = lib$addx (&QuadBytesCache,
                      &QuadBytesRaw,
                      &QuadBytesTotal, &Addx2);
   if (Debug) fprintf (stdout, "lib$addx() %%X%08.08X\n", status);

   FloatBytesCache = (float)QuadBytesCache[0] +
                     (float)QuadBytesCache[1] * (float)0xffffffff;
   if (Debug) fprintf (stdout, "FloatBytesCache: %f\n", FloatBytesCache);

   CommaNumber (64, &ProxyAccounting.QuadBytesCacheRx[0], BytesCacheRx); 
   CommaNumber (64, &ProxyAccounting.QuadBytesCacheTx[0], BytesCacheTx); 
   CommaNumber (64, &ProxyAccounting.QuadBytesRawRx[0], BytesRawRx); 
   CommaNumber (64, &ProxyAccounting.QuadBytesRawTx[0], BytesRawTx); 
   CommaNumber (64, &QuadBytesRaw[0], BytesTotal); 

   /********************/
   /* bytes percentage */
   /********************/

   FloatBytesNotCacheable = (float)QuadBytesNotCacheable[0] +
                            (float)QuadBytesNotCacheable[1] * (float)0xffffffff;
   if (Debug)
      fprintf (stdout, "FloatBytesNotCacheable: %f\n", FloatBytesNotCacheable);

   FloatBytesNetwork = (float)QuadBytesRaw[0] +
                   (float)QuadBytesRaw[1] * (float)0xffffffff;
   if (Debug) fprintf (stdout, "FloatBytesNetwork: %f\n", FloatBytesNetwork);

   FloatBytesTotal = (float)QuadBytesTotal[0] +
                     (float)QuadBytesTotal[1] * (float)0xffffffff;
   if (Debug) fprintf (stdout, "FloatBytesTotal: %f\n", FloatBytesTotal);

   if (FloatBytesTotal > 0.0)
   {
      PercentBytesCache = (int)(FloatPercent =
                          FloatBytesCache * 100.0 / FloatBytesTotal);
      if (FloatPercent - (float)PercentBytesCache >= 0.5) PercentBytesCache++;

      PercentBytesNotCacheable =
         (int)(FloatPercent = FloatBytesNotCacheable * 100.0 / FloatBytesTotal);
      if (FloatPercent - (float)PercentBytesNotCacheable >= 0.5)
         PercentBytesNotCacheable++;

      PercentBytesNetwork =
         (int)(FloatPercent = FloatBytesNetwork * 100.0 / FloatBytesTotal);
      if (FloatPercent - (float)PercentBytesNetwork >= 0.5)
         PercentBytesNetwork++;
   }
   else
      PercentBytesCache = PercentBytesNotCacheable = PercentBytesNetwork = 0;

   /*********************/
   /* counts percentage */
   /*********************/

   FloatCountRequestNotCacheable = ProxyAccounting.RequestNotCacheableCount;
   FloatCountResponseNotCacheable = ProxyAccounting.ResponseNotCacheableCount;
   FloatCountNetwork = ProxyAccounting.NetworkCount;
   FloatCountCacheRead = ProxyAccounting.CacheReadCount;
   FloatCountCacheWrite = ProxyAccounting.CacheWriteCount;
   FloatCountTotal = FloatCountNetwork + FloatCountCacheRead;

   if (FloatCountTotal > 0.0)
   {
      PercentCountNetwork = 
         (int)(FloatPercent = FloatCountNetwork * 100.0 / FloatCountTotal);
      if (FloatPercent - (float)PercentCountNetwork >= 0.5)
         PercentCountNetwork++;

      PercentCountCacheRead = (int)(FloatPercent =
                          FloatCountCacheRead * 100.0 / FloatCountTotal);
      if (FloatPercent - (float)PercentCountCacheRead >= 0.5)
         PercentCountCacheRead++;

      PercentCountRequestNotCacheable =
         (int)(FloatPercent =
               FloatCountRequestNotCacheable * 100.0 / FloatCountTotal);
      if (FloatPercent - (float)PercentCountRequestNotCacheable >= 0.5)
         PercentCountRequestNotCacheable++;
   }
   else
      PercentCountCacheRead = PercentCountResponseNotCacheable =
         PercentCountNetwork = 0;

   if (FloatCountNetwork > 0.0)
   {
      PercentCountCacheWrite = 
         (int)(FloatPercent = FloatCountCacheWrite * 100.0 / FloatCountNetwork);
      if (FloatPercent - (float)PercentCountCacheWrite >= 0.5)
         PercentCountCacheWrite++;

      PercentCountResponseNotCacheable =
         (int)(FloatPercent =
               FloatCountResponseNotCacheable * 100.0 / FloatCountNetwork);
      if (FloatPercent - (float)PercentCountResponseNotCacheable >= 0.5)
         PercentCountResponseNotCacheable++;
   }
   else
      PercentCountCacheWrite = PercentCountResponseNotCacheable = 0;

   /***********/
   /* display */
   /***********/

   vecptr = FaoVector;

   if (ProxyServingEnabled)
      *vecptr++ = "";
   else
      *vecptr++ =
" &nbsp;&nbsp;<FONT COLOR=\"#ff0000\">currently DISABLED</FONT>";

   *vecptr++ = ProxyAccounting.MethodGetCount;
   *vecptr++ = ProxyAccounting.MethodHeadCount;
   *vecptr++ = ProxyAccounting.MethodPostCount;
   *vecptr++ = ProxyAccounting.MethodPutCount;
   *vecptr++ = ProxyAccounting.MethodConnectCount;
   *vecptr++ = ProxyAccounting.MethodGetCount +
               ProxyAccounting.MethodHeadCount +
               ProxyAccounting.MethodPostCount +
               ProxyAccounting.MethodPutCount +
               ProxyAccounting.MethodConnectCount;
   *vecptr++ = BytesRawRx;
   *vecptr++ = BytesRawTx;
   *vecptr++ = PercentBytesNetwork;
   *vecptr++ = BytesCacheRx;
   *vecptr++ = BytesCacheTx;
   *vecptr++ = PercentBytesCache;
   *vecptr++ = BytesTotal;
   *vecptr++ = ProxyAccounting.CacheReadCount;
   *vecptr++ = ProxyAccounting.CacheRead304Count;
   *vecptr++ = PercentCountCacheRead;
   *vecptr++ = ProxyAccounting.CacheWriteCount;
   *vecptr++ = PercentCountCacheWrite;
   *vecptr++ = ProxyAccounting.RequestNotCacheableCount;
   *vecptr++ = PercentCountRequestNotCacheable;
   *vecptr++ = ProxyAccounting.ResponseNotCacheableCount;
   *vecptr++ = PercentCountResponseNotCacheable;
   *vecptr++ = PercentBytesNotCacheable;
   *vecptr++ = ProxyAccounting.NetworkCount;
   *vecptr++ = PercentCountNetwork;
   *vecptr++ = ProxyAccounting.LookupNoneCount;
   *vecptr++ = ProxyAccounting.LookupDnsCount;
   *vecptr++ = ProxyAccounting.LookupCacheCount;
   *vecptr++ = ProxyAccounting.LookupErrorCount;

   status = sys$faol (&ResponseFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF) return (status);

   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, 0, Buffer, Length);
   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
*/ 

ProxyMaintControl (struct RequestStruct *rqptr)

{
   register char  *cptr, *qptr, *sptr, *zptr;

   boolean  MakeChanges;
   int  status;
   unsigned short  Length;
   char  FieldName [128],
         FieldValue [256],
         ProxyCachePurgeList [64],
         ProxyCacheReloadList [64];


   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ProxyMaintControl()\n");

   MakeChanges = false;

   /**********************/
   /* parse content body */
   /**********************/

   qptr = rqptr->ContentBufferPtr;
   while (*qptr)
   {
      qptr = ParseQueryField (rqptr, qptr,
                              FieldName, sizeof(FieldName),
                              FieldValue, sizeof(FieldValue),
                              FI_LI);
      if (qptr == NULL)
      {
         /* error occured */
         rqptr->ResponseStatusCode = 400;
         ErrorGeneral (rqptr, "Query field problem.", FI_LI);
         return;
      }

      /************************/
      /* action button fields */
      /************************/

      if (strsame (FieldName, "hostname", -1))
      {
         ProxyResolveHostCache (NULL, NULL, 0);
         AdminControlSuccess (rqptr, "", "host name cache purged.");
      }
      else
      if (strsame (FieldName, "statistics", -1))
      {
         if (ProxyMaintScanBegin (PROXY_MAINT_SCAN_STATISTICS))
         AdminControlSuccess (rqptr, "",
"beginning statistics scan (results may take some time to be delivered).");
      }
      else
      if (strsame (FieldName, "routine", -1))
      {
         if (ProxyMaintScanBegin (PROXY_MAINT_SCAN_ROUTINE))
            AdminControlSuccess (rqptr, "", "begining routine purge.");
         else
         {
            rqptr->ResponseStatusCode = 400;
            ErrorGeneral (rqptr, "Routine purge not started!", FI_LI);
         }
         return;
      }
      else
      if (strsame (FieldName, "reactive", -1))
      {
         if (ProxyMaintScanBegin (PROXY_MAINT_SCAN_REACTIVE))
            AdminControlSuccess (rqptr, "", "begining reactive purge.");
         else
         {
            rqptr->ResponseStatusCode = 400;
            ErrorGeneral (rqptr, "Routine purge not started!", FI_LI);
         }
         return;
      }
      else
      if (strsame (FieldName, "makechanges", -1))
         MakeChanges = true;
      else

      /************************/
      /* "makechanges" fields */
      /************************/

      if (strsame (FieldName, "ProxyAddForwardedByEnabled", -1))
      {
          if (toupper(FieldValue[0]) == 'E')
            ProxyAddForwardedByEnabled = true;
         else
         if (toupper(FieldValue[0]) == 'D')
            ProxyAddForwardedByEnabled = false;
      }
      else
      if (strsame (FieldName, "ProxyServingEnabled", -1))
      {
          if (toupper(FieldValue[0]) == 'E')
            ProxyServingEnabled = ProxyAccounting.ServingEnabled = true;
         else
         if (toupper(FieldValue[0]) == 'D')
            ProxyServingEnabled = ProxyAccounting.ServingEnabled = false;
      }
      else
      if (strsame (FieldName, "ProxyHostLookupRetryCount", -1))
      {
         ProxyHostLookupRetryCount = atoi(FieldValue);
         if (ProxyHostLookupRetryCount < 0 || ProxyHostLookupRetryCount > 100)
            ProxyHostLookupRetryCount = 0;
      }
      else
      if (strsame (FieldName, "ProxyHostCachePurgeHours", -1))
      {
         ProxyHostCachePurgeHours = atoi(FieldValue);
         if (ProxyHostCachePurgeHours < 1 || ProxyHostCachePurgeHours > 168)
            ProxyHostCachePurgeHours  = 1;
      }
      else
      if (strsame (FieldName, "ProxyCacheFileKBytesMax", -1))
         ProxyCacheFileKBytesMax = atoi(FieldValue);
      else
      if (strsame (FieldName, "ProxyCacheReloadList", -1))
      {
         zptr = (sptr = ProxyCacheReloadList) + sizeof(ProxyCacheReloadList);
         for (cptr = FieldValue; *cptr && sptr < zptr; *sptr++ = *cptr++);
         if (sptr >= zptr)
         {
            ErrorGeneralOverflow (rqptr, FI_LI);
            return;
         }
         *sptr = '\0';
         ProxyCacheReloadListPtr = ProxyCacheReloadList;
      }
      else
      if (strsame (FieldName, "ProxyCacheRoutineHourOfDay", -1))
         ProxyCacheRoutineHourOfDay = atoi(FieldValue);
      else
      if (strsame (FieldName, "ProxyCacheDeviceCheckMinutes", -1))
         ProxyCacheDeviceCheckMinutes = atoi(FieldValue);
      else
      if (strsame (FieldName, "ProxyCacheDeviceMaxPercent", -1))
         ProxyCacheDeviceMaxPercent = atoi(FieldValue);
      else
      if (strsame (FieldName, "ProxyCacheDevicePurgePercent", -1))
         ProxyCacheDevicePurgePercent = atoi(FieldValue);
      else
      if (strsame (FieldName, "ProxyCachePurgeList", -1))
      {
         zptr = (sptr = ProxyCachePurgeList) + sizeof(ProxyCachePurgeList);
         for (cptr = FieldValue; *cptr && sptr < zptr; *sptr++ = *cptr++);
         if (sptr >= zptr)
         {
            ErrorGeneralOverflow (rqptr, FI_LI);
            return;
         }
         *sptr = '\0';
         ProxyCachePurgeListPtr = ProxyCachePurgeList;
      }
      else
      if (strsame (FieldName, "ProxyReportLog", -1))
      {
          if (toupper(FieldValue[0]) == 'E')
            ProxyReportLog = true;
         else
         if (toupper(FieldValue[0]) == 'D')
            ProxyReportLog = false;
      }
      else
      if (strsame (FieldName, "ProxyReportCacheLog", -1))
      {
          if (toupper(FieldValue[0]) == 'E')
            ProxyReportCacheLog = true;
         else
         if (toupper(FieldValue[0]) == 'D')
            ProxyReportCacheLog = false;
      }
      else
      {
         /***********/
         /* unknown */
         /***********/

         rqptr->ResponseStatusCode = 400;
         ErrorGeneral (rqptr, "Unknown query field.", FI_LI);
         return;
      }
   }

   if (MakeChanges)
   {
      ProxyCacheInitValues ();
      ProxyMaintInit ();
      /* set logical names used by HTTPd monitor utility */
      if (MonitorEnabled) DefineMonitorLogicals (NULL);
      AdminControlSuccess (rqptr, "", "new proxy values loaded.");
   }
}

/****************************************************************************/
